from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import math
from collections import OrderedDict

import veriloggen as vg

import nngen.basic_types as bt
import nngen.util as util

from . import basic
from . import concat


class normalize(basic.multiply_add_rshift_clip):
    """ for Batchnorm """
    pass


class scaled_add(bt._ElementwiseOperator):
    """ for ResBlock """
    input_chainable = True
    output_chainable = True

    def __sub_str__(self):
        a_scale = ' a_scale:%d' % self.a_scale
        b_scale = ' b_scale:%d' % self.b_scale
        shamt = ' shamt:%d' % self.shamt
        ret = [a_scale, b_scale, shamt]
        return ''.join(ret)

    def get_local_control_param_values(self):
        return OrderedDict([('a_scale_cparam', int(self.a_scale)),
                            ('b_scale_cparam', int(self.b_scale)),
                            ('shamt_cparam', self.shamt)])

    def op(self, strm, *args, **kwargs):
        a_datawidth = self.args[0].get_op_width()
        a_point = self.args[0].get_op_point()
        a_signed = self.args[0].get_signed()
        b_datawidth = self.args[1].get_op_width()
        b_point = self.args[1].get_op_point()
        b_signed = self.args[1].get_signed()

        a_scale = strm.ReinterpretCast(self.a_scale_cparam,
                                       width=a_datawidth,
                                       signed=a_signed)
        b_scale = strm.ReinterpretCast(self.b_scale_cparam,
                                       width=b_datawidth,
                                       signed=b_signed)

        mul = strm.Times(args[0], a_scale)

        if self.sum_dtype is not None:
            mul.width = self.sum_dtype.width
            mul.signed = self.sum_dtype.signed
        else:
            mul.width = a_datawidth + vg.get_width(self.a_scale_cparam)
            mul.signed = self.dtype.signed

        if self.sum_dtype is not None and mul.point != self.sum_dtype.point:
            mul = strm.Cast(mul, point=self.sum_dtype.point)

        madd = strm.Madd(args[1], b_scale, mul)

        if self.sum_dtype is not None:
            madd.width = self.sum_dtype.width
            madd.signed = self.sum_dtype.signed
        else:
            madd.width = max(b_datawidth + vg.get_width(self.b_scale_cparam), mul.width)
            madd.signed = self.dtype.signed

        if self.sum_dtype is not None and madd.point != self.sum_dtype.point:
            madd = strm.Cast(madd, point=self.sum_dtype.point)

        shamt = strm.ReinterpretCast(self.shamt_cparam,
                                     width=self.shamt_cparam.width,
                                     signed=False)
        sra = strm.Sra(madd, shamt)

        p_th, n_th = util.clip_threshold(self.dtype.width, self.dtype.signed, self.asymmetric_clip)
        p = strm.Mux(sra > p_th, p_th, sra)
        n = strm.Mux(sra < n_th, n_th, sra)

        return strm.Mux(sra >= 0, p, n)

    def __init__(self, a, b, a_scale, b_scale, shamt,
                 asymmetric_clip=False,
                 dtype=None, sum_dtype=None, name=None, par=1):

        shape = None
        self.a_scale = a_scale
        self.b_scale = b_scale
        self.shamt = shamt
        bt._ElementwiseOperator.__init__(self, a, b,
                                         dtype=dtype, shape=shape, name=name, par=par)
        self.asymmetric_clip = asymmetric_clip
        self.sum_dtype = sum_dtype

    def eval(self, memo, input_dict, **kwargs):
        kwargs['a_scale'] = self.a_scale
        kwargs['b_scale'] = self.b_scale
        kwargs['shamt'] = self.shamt
        kwargs['asymmetric_clip'] = self.asymmetric_clip
        kwargs['sum_dtype'] = self.sum_dtype
        kwargs['a_dtype'] = self.args[0].dtype
        kwargs['b_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class scaled_concat(concat):

    def __sub_str__(self):
        base = concat.__sub_str__(self)
        scales = ' scales:%s' % str(tuple(self.scales))
        shamt = ' shamt:%d' % self.shamt
        return ''.join([base, scales, shamt])

    def get_control_param_values(self):
        ret = concat.get_control_param_values(self)

        buffered = True
        ret['buffered'] = buffered

        # for __str__
        self.buffered_value = buffered

        return ret

    def get_local_control_param_values(self):
        return OrderedDict([('scale_cparams', [int(scale) for scale in self.scales]),
                            ('shamt_cparam', self.shamt)])

    def __init__(self, values, scales, shamt, axis,
                 asymmetric_clip=False,
                 dtype=None, mul_dtype=None, name=None):
        if not isinstance(scales, (tuple, list)):
            raise TypeError('scales must be tuple or list.')

        # for quantization, scales must be list
        self.scales = list(scales)
        self.shamt = shamt
        concat.__init__(self, values, axis, dtype, name)
        self.asymmetric_clip = asymmetric_clip
        self.mul_dtype = mul_dtype

    def get_stream_func(self):
        def func(strm):
            datawidth = self.args[0].get_op_width()

            src = strm.source(datawidth=datawidth)
            sel = strm.parameter(datawidth=bt.log_width(len(self.scale_cparams)),
                                 signed=False)

            scale_width = max(*[scale.width for scale in self.scale_cparams])
            scale_signed = max(*[1 if scale.signed else 0 for scale in self.scale_cparams]) == 1
            selected_scale = strm.ReinterpretCast(self.scale_cparams[-1],
                                                  width=scale_width,
                                                  signed=scale_signed)

            for i, scale in enumerate(reversed(self.scale_cparams[:-1])):
                scale = strm.ReinterpretCast(scale,
                                             width=scale_width,
                                             signed=scale_signed)
                selected_scale = strm.Mux(sel == len(self.scale_cparams) - 2 - i,
                                          scale, selected_scale)
                selected_scale.width = scale_width
                selected_scale.signed = scale_signed
                selected_scale.latency = 0

            mul = strm.Times(src, selected_scale)

            if self.mul_dtype is not None:
                mul.width = self.mul_dtype.width
                mul.signed = self.mul_dtype.signed
            else:
                mul.width = datawidth + scale_width
                mul.signed = self.dtype.signed

            if self.mul_dtype is not None and mul.point != self.mul_dtype.point:
                mul = strm.Cast(mul, point=self.mul_dtype.point)

            shamt = strm.ReinterpretCast(self.shamt_cparam,
                                         width=self.shamt_cparam.width,
                                         signed=False)
            sra = strm.Sra(mul, shamt)

            p_th, n_th = util.clip_threshold(
                self.dtype.width, self.dtype.signed, self.asymmetric_clip)
            p = strm.Mux(sra > p_th, p_th, sra)
            n = strm.Mux(sra < n_th, n_th, sra)
            dst = strm.Mux(sra >= 0, p, n)

            strm.sink(dst)

        return func

    def control_sequence(self, fsm):
        arg_gaddrs = [self.m.Reg(self._name('arg_gaddr_%d' % i),
                                 self.maxi.addrwidth, initval=0)
                      for i, _ in enumerate(self.arg_objaddrs)]
        out_gaddr = self.m.Reg(self._name('out_gaddr'),
                               self.maxi.addrwidth, initval=0)

        arg_laddr = self.m.Reg(self._name('arg_laddr'),
                               self.maxi.addrwidth, initval=0)
        copy_laddr = self.m.Reg(self._name('copy_laddr'),
                                self.maxi.addrwidth, initval=0)
        copy_size = self.m.Reg(self._name('copy_size'),
                               self.maxi.addrwidth, initval=0)
        sum_read_sizes = self.m.Wire(self._name('sum_read_sizes'),
                                     max([i.width for i in self.arg_read_sizes])
                                     + int(math.ceil(math.log2(len(self.arg_read_sizes)))))
        sum_read_sizes.assign(vg.Add(*self.arg_read_sizes))

        out_addr_inc_unbuffered = self.m.Reg(self._name('out_addr_inc_unbuffered'),
                                             self.maxi.addrwidth, initval=0)

        arg_select = self.m.Reg(self._name('arg_select'),
                                int(max(math.ceil(math.log(len(self.args), 2)), 1)),
                                initval=0)
        prev_arg_select = self.m.Reg(self._name('prev_arg_select'),
                                     int(max(math.ceil(math.log(len(self.args), 2)), 1)),
                                     initval=0)
        arg_chunk_count = self.m.Reg(self._name('arg_chunk_count'),
                                     self.maxi.addrwidth + 1, initval=0)
        out_count = self.m.Reg(self._name('out_count'),
                               self.maxi.addrwidth + 1, initval=0)

        # --------------------
        # initialization phase
        # --------------------
        fsm(
            [arg_gaddr(0) for arg_gaddr in arg_gaddrs],
            out_gaddr(0),
            arg_laddr(0),
            copy_laddr(0),
            copy_size(0),
            out_addr_inc_unbuffered(0),
            arg_select(0),
            prev_arg_select(0),
            arg_chunk_count(0),
            out_count(0)
        )

        fsm.goto_next()

        # --------------------
        # Read phase
        # --------------------
        state_read = fsm.current

        fsm.inc()

        state_read_begin_list = []
        state_read_end_list = []

        for (arg,
             arg_objaddr, arg_gaddr, arg_addr_inc,
             arg_read_size, arg_chunk_size) in zip(
                 self.args,
                 self.arg_objaddrs, arg_gaddrs, self.arg_addr_incs,
                 self.arg_read_sizes, self.arg_chunk_sizes):

            b = fsm.current
            state_read_begin_list.append(b)

            # normal
            laddr = arg_laddr
            gaddr = arg_objaddr + arg_gaddr

            bt.bus_lock(self.maxi, fsm)
            bt.dma_read(self.maxi, fsm, self.input_rams[0], laddr, gaddr, arg_read_size)
            bt.bus_unlock(self.maxi, fsm)

            fsm(
                arg_gaddr.add(arg_addr_inc),
                arg_laddr.add(arg_read_size),
                arg_chunk_count.inc(),
                copy_size(arg_read_size),
                out_addr_inc_unbuffered(arg_addr_inc)
            )
            fsm.If(arg_chunk_count == arg_chunk_size - 1)(
                arg_chunk_count(0),
                arg_select.inc()
            )
            fsm.If(arg_chunk_count == arg_chunk_size - 1,
                   arg_select == len(self.args) - 1)(
                arg_select(0)
            )
            fsm(
                prev_arg_select(arg_select)
            )

            e = fsm.current
            state_read_end_list.append(e)

            fsm.inc()

        state_read_end = fsm.current

        for i, b in enumerate(state_read_begin_list):
            fsm.If(arg_select == i).goto_from(state_read, b)

        for i, e in enumerate(state_read_end_list):
            fsm.goto_from(e, state_read_end)

        # --------------------
        # Copy phase
        # --------------------
        state_copy = fsm.current

        name = list(self.stream.sources.keys())[0]
        self.stream.set_source(fsm, name, self.input_rams[0], 0, copy_size)
        fsm.set_index(fsm.current - 1)

        name = list(self.stream.parameters.keys())[0]
        self.stream.set_parameter(fsm, name, prev_arg_select)
        fsm.set_index(fsm.current - 1)

        name = list(self.stream.sinks.keys())[0]
        self.stream.set_sink(fsm, name, self.output_rams[0], copy_laddr, copy_size)
        self.stream.run(fsm)
        self.stream.join(fsm)

        fsm(
            arg_laddr(0),
            copy_laddr.add(copy_size)
        )
        fsm.goto_next()

        fsm.If(copy_laddr < sum_read_sizes).goto(state_read)
        fsm.If(copy_laddr >= sum_read_sizes).goto_next()

        state_copy_end = fsm.current
        fsm.If(vg.Not(self.buffered)).goto_from(state_copy, state_copy_end)

        # --------------------
        # Write phase
        # --------------------
        state_write = fsm.current
        fsm.inc()

        # Case with Copy
        state_write_buffered = fsm.current

        laddr = 0
        gaddr = self.objaddr + out_gaddr
        bt.bus_lock(self.maxi, fsm)
        bt.dma_write(self.maxi, fsm,
                     self.output_rams[0], laddr, gaddr, self.out_write_size)
        bt.bus_unlock(self.maxi, fsm)

        fsm(
            copy_laddr(0),
            out_gaddr.add(self.out_addr_inc),
            out_count.inc()
        )

        state_write_end_buffered = fsm.current
        fsm.inc()

        # Case without Copy
        state_write_unbuffered = fsm.current

        laddr = 0
        gaddr = self.objaddr + out_gaddr
        bt.bus_lock(self.maxi, fsm)
        bt.dma_write(self.maxi, fsm,
                     self.input_rams[0], laddr, gaddr, copy_size)
        bt.bus_unlock(self.maxi, fsm)

        fsm(
            arg_laddr(0),
            out_gaddr.add(out_addr_inc_unbuffered),
            out_count.inc()
        )

        state_write_end_unbuffered = fsm.current
        fsm.inc()

        state_write_end = fsm.current

        fsm.If(self.buffered).goto_from(state_write, state_write_buffered)
        fsm.If(vg.Not(self.buffered)).goto_from(state_write, state_write_unbuffered)

        fsm.goto_from(state_write_end_buffered, state_write_end)
        fsm.goto_from(state_write_end_unbuffered, state_write_end)

        # --------------------
        # update for next iteration
        # --------------------
        fsm.If(out_count < self.num_steps).goto(state_read)
        fsm.If(out_count == self.num_steps).goto_next()

    def eval(self, memo, input_dict, **kwargs):
        kwargs['scales'] = self.scales
        kwargs['asymmetric_clip'] = self.asymmetric_clip
        kwargs['shamt'] = self.shamt
        kwargs['asymmetric_clip'] = self.asymmetric_clip
        kwargs['mul_dtype'] = self.mul_dtype
        return concat.eval(self, memo, input_dict, **kwargs)


class scaled_multiply(bt._ElementwiseOperator):

    input_chainable = True
    output_chainable = True

    def __sub_str__(self):
        shamt = ' shamt:%d' % self.shamt
        ret = [shamt]
        return ''.join(ret)

    def get_local_control_param_values(self):
        return OrderedDict([('shamt_cparam', self.shamt)])

    def op(self, strm, *args, **kwargs):
        a_datawidth = self.args[0].get_op_width()
        a_point = self.args[0].get_op_point()
        a_signed = self.args[0].get_signed()
        b_datawidth = self.args[1].get_op_width()
        b_point = self.args[1].get_op_point()
        b_signed = self.args[1].get_signed()

        mul = strm.Times(args[0], args[1])

        if self.mul_dtype is not None:
            mul.width = self.mul_dtype.width
            mul.signed = self.mul_dtype.signed
        else:
            mul.width = a_datawidth + b_datawidth
            mul.signed = self.dtype.signed

        if self.mul_dtype is not None and mul.point != self.mul_dtype.point:
            mul = strm.Cast(mul, point=self.mul_dtype.point)

        shamt = strm.ReinterpretCast(self.shamt_cparam,
                                     width=self.shamt_cparam.width,
                                     signed=False)
        sra = strm.Sra(mul, shamt)

        p_th, n_th = util.clip_threshold(self.dtype.width, self.dtype.signed, self.asymmetric_clip)
        p = strm.Mux(sra > p_th, p_th, sra)
        n = strm.Mux(sra < n_th, n_th, sra)
        return strm.Mux(sra >= 0, p, n)

    def __init__(self, a, b, shamt,
                 asymmetric_clip=False,
                 dtype=None, mul_dtype=None, name=None, par=1):

        shape = None
        self.shamt = shamt
        bt._ElementwiseOperator.__init__(self, a, b,
                                         dtype=dtype, shape=shape, name=name, par=par)
        self.asymmetric_clip = asymmetric_clip
        self.mul_dtype = mul_dtype

    def eval(self, memo, input_dict, **kwargs):
        kwargs['shamt'] = self.shamt
        kwargs['asymmetric_clip'] = self.asymmetric_clip
        kwargs['mul_dtype'] = self.mul_dtype
        kwargs['a_dtype'] = self.args[0].dtype
        kwargs['b_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class scaled_div(bt._ElementwiseOperator):

    input_chainable = True
    output_chainable = True

    def __sub_str__(self):
        shamt = ' shamt:%d' % self.shamt
        ret = [shamt]
        return ''.join(ret)

    def get_local_control_param_values(self):
        return OrderedDict([('shamt_cparam', self.shamt)])

    def op(self, strm, *args, **kwargs):
        a_datawidth = self.args[0].get_op_width()
        a_point = self.args[0].get_op_point()
        a_signed = self.args[0].get_signed()
        b_datawidth = self.args[1].get_op_width()
        b_point = self.args[1].get_op_point()
        b_signed = self.args[1].get_signed()

        shamt = strm.ReinterpretCast(self.shamt_cparam,
                                     width=self.shamt_cparam.width,
                                     signed=False)

        sll = strm.Sll(args[0], shamt)
        sll.width = a_datawidth + b_datawidth
        div = strm.Div(sll, args[1])

        if self.div_dtype is not None:
            div.width = self.div_dtype.width
            div.signed = self.div_dtype.signed
        else:
            div.signed = self.dtype.signed

        if self.div_dtype is not None and div.point != self.div_dtype.point:
            div = strm.Cast(div, point=self.div_dtype.point)

        p_th, n_th = util.clip_threshold(self.dtype.width, self.dtype.signed, self.asymmetric_clip)
        p = strm.Mux(div > p_th, p_th, div)
        n = strm.Mux(div < n_th, n_th, div)
        return strm.Mux(div >= 0, p, n)

    def __init__(self, a, b, shamt,
                 asymmetric_clip=False,
                 dtype=None, div_dtype=None, name=None, par=1):

        shape = None
        self.shamt = shamt
        bt._ElementwiseOperator.__init__(self, a, b,
                                         dtype=dtype, shape=shape, name=name, par=par)
        self.asymmetric_clip = asymmetric_clip
        self.div_dtype = div_dtype

    def eval(self, memo, input_dict, **kwargs):
        kwargs['shamt'] = self.shamt
        kwargs['asymmetric_clip'] = self.asymmetric_clip
        kwargs['div_dtype'] = self.div_dtype
        kwargs['a_dtype'] = self.args[0].dtype
        kwargs['b_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np

import nngen.util as util


def matmul(a, b,
           bias=None, scale=None,
           transposed_a=False, transposed_b=True,
           rshift_mul=None, rshift_sum=None, rshift_out=None,
           act_func=None,
           asymmetric_clip=False,
           dtype=None, mul_dtype=None, sum_dtype=None,
           name=None,
           par_left_col=1, par_left_row=1, par_out_col=1,
           concur_out_col=None, stationary='right',
           left_ram_size=None, right_ram_size=None,
           bias_ram_size=None, scale_ram_size=None,
           vshamt_mul_ram_size=None, vshamt_sum_ram_size=None, vshamt_out_ram_size=None,
           out_ram_size=None,
           disable_keep_left=False,
           a_dtype=None, b_dtype=None,
           bias_dtype=None, scale_dtype=None):

    if transposed_a:
        a = a.transpose()

    if not transposed_b:
        b = b.transpose()

    if a.shape[1] != b.shape[1]:
        raise ValueError("shape mismatch: %s != %s" %
                         (str(a.shape), str(b.shape)))

    c_shape = (a.shape[0], b.shape[0])

    if bias is None:
        bias = np.zeros([c_shape[-1]], dtype=np.int64)
    elif not isinstance(bias, np.ndarray):
        new_bias = np.zeros([c_shape[-1]], dtype=np.int64)
        for i in range(new_bias.shape[-1]):
            new_bias[i] = bias
        bias = new_bias
    elif len(bias.shape) == 1 and bias.shape[0] == 1:
        new_bias = np.zeros([c_shape[-1]], dtype=np.int64)
        for i in range(new_bias.shape[-1]):
            new_bias[i] = bias[0]
        bias = new_bias

    if scale is None:
        scale = np.ones([c_shape[-1]], dtype=np.int64)
    elif not isinstance(scale, np.ndarray):
        new_scale = np.zeros([c_shape[-1]], dtype=np.int64)
        for i in range(new_scale.shape[-1]):
            new_scale[i] = scale
        scale = new_scale
    elif len(scale.shape) == 1 and scale.shape[0] == 1:
        new_scale = np.zeros([c_shape[-1]], dtype=np.int64)
        for i in range(new_scale.shape[-1]):
            new_scale[i] = scale[0]
        scale = new_scale

    if rshift_mul is None:
        rshift_mul = np.zeros([c_shape[-1]], dtype=np.int64)
    elif not isinstance(rshift_mul, np.ndarray):
        new_rshift_mul = np.zeros([c_shape[-1]], dtype=np.int64)
        for i in range(new_rshift_mul.shape[-1]):
            new_rshift_mul[i] = rshift_mul
        rshift_mul = new_rshift_mul
    elif len(rshift_mul.shape) == 1 and rshift_mul.shape[0] == 1:
        new_rshift_mul = np.zeros([c_shape[-1]], dtype=np.int64)
        for i in range(new_rshift_mul.shape[-1]):
            new_rshift_mul[i] = rshift_mul[0]
        rshift_mul = new_rshift_mul

    if rshift_sum is None:
        rshift_sum = np.zeros([c_shape[-1]], dtype=np.int64)
    elif not isinstance(rshift_sum, np.ndarray):
        new_rshift_sum = np.zeros([c_shape[-1]], dtype=np.int64)
        for i in range(new_rshift_sum.shape[-1]):
            new_rshift_sum[i] = rshift_sum
        rshift_sum = new_rshift_sum
    elif len(rshift_sum.shape) == 1 and rshift_sum.shape[0] == 1:
        new_rshift_sum = np.zeros([c_shape[-1]], dtype=np.int64)
        for i in range(new_rshift_sum.shape[-1]):
            new_rshift_sum[i] = rshift_sum[0]
        rshift_sum = new_rshift_sum

    if rshift_out is None:
        rshift_out = np.zeros([c_shape[-1]], dtype=np.int64)
    elif not isinstance(rshift_out, np.ndarray):
        new_rshift_out = np.zeros([c_shape[-1]], dtype=np.int64)
        for i in range(new_rshift_out.shape[-1]):
            new_rshift_out[i] = rshift_out
        rshift_out = new_rshift_out
    elif len(rshift_out.shape) == 1 and rshift_out.shape[0] == 1:
        new_rshift_out = np.zeros([c_shape[-1]], dtype=np.int64)
        for i in range(new_rshift_out.shape[-1]):
            new_rshift_out[i] = rshift_out[0]
        rshift_out = new_rshift_out

    rshift_mul_pow = np.where(rshift_mul > np.zeros_like(rshift_mul, dtype=np.int64),
                              rshift_mul - 1,
                              np.zeros_like(rshift_mul))
    rshift_mul_round = np.where(rshift_mul > np.zeros_like(rshift_mul, dtype=np.int64),
                                np.power(np.ones_like(rshift_mul, dtype=np.int64) * 2,
                                         rshift_mul_pow),
                                np.zeros_like(rshift_mul, dtype=np.int64))

    rshift_sum_pow = np.where(rshift_sum > np.zeros_like(rshift_sum, dtype=np.int64),
                              rshift_sum - 1,
                              np.zeros_like(rshift_sum))
    rshift_sum_round = np.where(rshift_sum > np.zeros_like(rshift_sum, dtype=np.int64),
                                np.power(np.ones_like(rshift_sum, dtype=np.int64) * 2,
                                         rshift_sum_pow),
                                np.zeros_like(rshift_sum, dtype=np.int64))

    rshift_out_pow = np.where(rshift_out > np.zeros_like(rshift_out, dtype=np.int64),
                              rshift_out - 1,
                              np.zeros_like(rshift_out))
    rshift_out_round = np.where(rshift_out > np.zeros_like(rshift_out, dtype=np.int64),
                                np.power(np.ones_like(rshift_out, dtype=np.int64) * 2,
                                         rshift_out_pow),
                                np.zeros_like(rshift_out, dtype=np.int64))

    a_point = 0 if a_dtype is None else a_dtype.point
    b_point = 0 if b_dtype is None else b_dtype.point
    bias_point = 0 if bias_dtype is None else bias_dtype.point
    scale_point = 0 if scale_dtype is None else scale_dtype.point
    c_point = max(a_point, b_point) if dtype is None else dtype.point
    c_width = 32 if dtype is None else dtype.width

    mul_point = max(a_point, b_point)
    mul_shift = min(a_point, b_point)
    sum_point = mul_point
    add_point = max(sum_point, bias_point)

    sum_shift = add_point - sum_point
    bias_shift = add_point - bias_point
    shifted_bias = np.left_shift(bias, bias_shift)

    scl_point = max(sum_point, scale_point)
    scl_shift = min(sum_point, scl_point)
    shifted_scale = np.right_shift(scale, scl_shift)

    p_th = (1 << (c_width - 1)) - 1
    if asymmetric_clip:
        n_th = -1 * p_th - 1
    else:
        n_th = -1 * p_th

    p_th = p_th >> c_point
    n_th = n_th >> c_point

    def my_matmul_by_matmul(a, w):
        return np.matmul(a, w.T)

    def my_matmul_by_multiply(a, w):
        v = a.reshape([a.shape[0], 1, a.shape[1]])
        mul = np.multiply(v, w)
        mul = np.add(mul, rshift_mul_round.reshape([rshift_mul_round.shape[-1], 1]))
        mul = np.right_shift(mul, mul_shift)
        mul = np.right_shift(mul, rshift_mul.reshape([rshift_mul.shape[-1], 1]))
        return np.add.reduce(mul, axis=2)

    if mul_shift == 0 and rshift_mul_round.all() == 0 and rshift_mul.all() == 0:
        my_matmul = my_matmul_by_matmul
    else:
        my_matmul = my_matmul_by_multiply

    if act_func is None:
        def act_op(x): return x
    else:
        act_op = act_func.get_act_func()

    sum = my_matmul(a, b)

    sum = np.left_shift(sum, sum_shift)
    sum = np.add(sum, rshift_sum_round)
    sum = np.right_shift(sum, rshift_sum)
    sum = np.add(sum, shifted_bias)
    sum = np.multiply(sum, shifted_scale)
    frac = np.where(rshift_out != 0, np.where(sum >= 0, rshift_out_round, rshift_out_round - 1),
                    np.zeros_like(rshift_out, dtype=np.int64))
    sum = np.add(sum, frac)
    sum = np.right_shift(sum, rshift_out)
    sum = np.where(sum > p_th, p_th, np.where(sum < n_th, n_th, sum))

    c = act_op(sum)

    return c



from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import math
import numpy as np
import PIL
import urllib

import torch
import torchvision
import torchvision.transforms as transforms
import torch.nn as nn
import torch.nn.functional as F
import torch.autograd

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi


def run(act_dtype=ng.int16, weight_dtype=ng.int8,
        bias_dtype=ng.int32, scale_dtype=ng.int8,
        disable_fusion=False,
        conv2d_par_ich=1, conv2d_par_och=1, conv2d_par_col=1, conv2d_par_row=1,
        conv2d_concur_och=None, conv2d_stationary='filter',
        pool_par=1, elem_par=1,
        chunk_size=64, axi_datawidth=32, silent=False,

        onnx_filename='yolov3-tiny.onnx',
        weight_filename='yolov3-tiny.npz',
        verilog_filename=None,
        sim_filename=None,
        # simtype=None,  # no RTL simulation
        # simtype='iverilog',
        simtype='verilator',

        cfg_filename='yolov3-tiny.cfg',
        weights_filename='yolov3-tiny.weights',
        model_path='yolov3'):

    # input mean and standard deviation
    imagenet_mean = np.array([0.485, 0.456, 0.406]).astype(np.float32)
    imagenet_std = np.array([0.229, 0.224, 0.225]).astype(np.float32)

    img_size = (416, 416)
    act_shape = (1, img_size[0], img_size[1], 3)

    # pytorch model
    model_url = "https://github.com/ultralytics/yolov3"
    if not os.path.isdir(model_path):
        raise FileNotFoundError("Download the YOLOv3 model using Pytorch, such as "
                                "'%s'. Then extract it, and rename it as '%s'" %
                                (model_url, model_path))

    # Darknet model configuration and pretrained weights
    cfg_url = "https://github.com/pjreddie/darknet/blob/master/cfg/yolov3-tiny.cfg"
    if not os.path.isfile(cfg_filename):
        urllib.request.urlretrieve(cfg_url, cfg_filename)

    weights_url = "https://pjreddie.com/media/files/yolov3-tiny.weights"
    if not os.path.isfile(weights_filename):
        urllib.request.urlretrieve(weights_url, weights_filename)

    sys.path.insert(0, model_path)
    import models
    models.ONNX_EXPORT = True

    model = models.Darknet(cfg_filename, img_size).to('cpu')
    models.load_darknet_weights(model, weights_filename)

    # Pytorch to ONNX
    dummy_input = torch.randn(*act_shape).transpose(1, 3)
    input_names = ['act']
    output_names = ['scores', 'boxes']
    model.eval()
    torch.onnx.export(model, dummy_input, onnx_filename,
                      input_names=input_names, output_names=output_names)

    # --------------------
    # (1) Represent a DNN model as a dataflow by NNgen operators
    # --------------------

    # ONNX to NNgen
    dtypes = {}
    shapes = {}
    (outputs, placeholders, variables,
     constants, operators) = ng.from_onnx(onnx_filename,
                                          value_dtypes=dtypes,
                                          value_shapes=shapes,
                                          default_placeholder_dtype=act_dtype,
                                          default_variable_dtype=weight_dtype,
                                          default_constant_dtype=weight_dtype,
                                          default_operator_dtype=act_dtype,
                                          default_scale_dtype=scale_dtype,
                                          default_bias_dtype=bias_dtype,
                                          disable_fusion=disable_fusion,
                                          verbose=False)

    # --------------------
    # (2) Assign quantized weights to the NNgen operators
    # --------------------

    if act_dtype.width > 8:
        act_scale_factor = 128
    else:
        act_scale_factor = int(round(2 ** (act_dtype.width - 1) * 0.5))

    input_scale_factors = {'act': act_scale_factor}
    input_means = {'act': imagenet_mean * act_scale_factor}
    input_stds = {'act': imagenet_std * act_scale_factor}

    ng.quantize(outputs, input_scale_factors, input_means, input_stds)

    # --------------------
    # (3) Assign hardware attributes
    # --------------------

    for op in operators.values():
        if isinstance(op, ng.conv2d):
            op.attribute(par_ich=conv2d_par_ich,
                         par_och=conv2d_par_och,
                         par_col=conv2d_par_col,
                         par_row=conv2d_par_row,
                         concur_och=conv2d_concur_och,
                         stationary=conv2d_stationary)

        if isinstance(op, (ng.avg_pool, ng.max_pool,
                           ng.avg_pool_serial, ng.max_pool_serial)):
            op.attribute(par=pool_par)

        if ng.is_elementwise_operator(op):
            op.attribute(par=elem_par)

    # --------------------
    # (4) Verify the DNN model behavior by executing the NNgen dataflow as a software
    # --------------------

    act = placeholders['act']
    outs = (outputs['scores'], outputs['boxes'])

    # verification data
    img = np.array(PIL.Image.open('car416x416.png').convert('RGB')).astype(np.float32)
    img = img.reshape([1] + list(img.shape))

    img = img / 255
    img = (img - imagenet_mean) / imagenet_std

    # execution on pytorch
    model_input = img

    if act.perm is not None:
        model_input = np.transpose(model_input, act.reversed_perm)

    model.eval()
    model_rslts = model(torch.from_numpy(model_input))
    model_outs = [rslt.detach().numpy() for rslt in model_rslts]
    model_outs = [(np.transpose(model_out, act.perm)
                   if act.perm is not None and len(model_out.shape) == len(act.shape)
                   else model_out)
                  for model_out in model_outs]
    scaled_model_outs = [model_out * out.scale_factor
                         for model_out, out in zip(model_outs, outs)]

    # software-based verification
    vact = img * act_scale_factor
    vact = np.clip(vact,
                   -1.0 * (2 ** (act.dtype.width - 1) - 1),
                   1.0 * (2 ** (act.dtype.width - 1) - 1))
    vact = np.round(vact).astype(np.int64)

    # compare outputs of hidden layers
    leaky_relu_ops = [v for k, v in operators.items()
                      if (isinstance(v, ng.conv2d) and
                          isinstance(v.act_func, ng.leaky_relu_base))]
    leaky_relu_ops = list(sorted(set(leaky_relu_ops), key=leaky_relu_ops.index))

    conv2d_ops = [v for k, v in operators.items()
                  if (isinstance(v, ng.conv2d) and v.act_func is None)]
    conv2d_ops = list(sorted(set(conv2d_ops), key=conv2d_ops.index))

    # only 1st output
    sub_ops = leaky_relu_ops[:9] + conv2d_ops[:1]
    sub_outs = ng.eval(sub_ops, act=vact)
    sub_outs = [sub_out.transpose([0, 3, 1, 2]) for sub_out in sub_outs]
    sub_scale_factors = [sub_op.scale_factor for sub_op in sub_ops]

    model.eval()
    mouts = []
    # all Conv2d-LeakyReLU layers before YOLOLayer
    mouts.append(nn.Sequential(model.module_list[0])(
        torch.from_numpy(model_input)).detach().numpy())
    mouts.append(nn.Sequential(*model.module_list[0:3])(
        torch.from_numpy(model_input)).detach().numpy())
    mouts.append(nn.Sequential(*model.module_list[0:5])(
        torch.from_numpy(model_input)).detach().numpy())
    mouts.append(nn.Sequential(*model.module_list[0:7])(
        torch.from_numpy(model_input)).detach().numpy())
    mouts.append(nn.Sequential(*model.module_list[0:9])(
        torch.from_numpy(model_input)).detach().numpy())
    mouts.append(nn.Sequential(*model.module_list[0:11])(
        torch.from_numpy(model_input)).detach().numpy())
    mouts.append(nn.Sequential(*model.module_list[0:13])(
        torch.from_numpy(model_input)).detach().numpy())
    mouts.append(nn.Sequential(*model.module_list[0:14])(
        torch.from_numpy(model_input)).detach().numpy())
    mouts.append(nn.Sequential(*model.module_list[0:15])(
        torch.from_numpy(model_input)).detach().numpy())
    mouts.append(nn.Sequential(*model.module_list[0:16])(
        torch.from_numpy(model_input)).detach().numpy())

    scaled_mouts = [mout * scale_factor
                    for mout, scale_factor in zip(mouts, sub_scale_factors)]

    sub_mean_square_errors = [np.sum((sub_out - mout) ** 2) / sub_out.size
                              for mout, sub_out in zip(scaled_mouts, sub_outs)]
    sub_corrcoefs = [np.corrcoef(mout.reshape([-1]), sub_out.reshape([-1]))
                     for mout, sub_out in zip(mouts, sub_outs)]

    # compare prediction results
    vouts = ng.eval(outs, act=vact)

    mean_square_errors = [np.sum((vout - scaled_model_out) ** 2) / vout.size
                          for vout, scaled_model_out in zip(vouts, scaled_model_outs)]
    corrcoefs = [np.corrcoef(model_out.reshape([-1]), vout.reshape([-1]))
                 for model_out, vout in zip(model_outs, vouts)]

    # breakpoint()

    # --------------------
    # (5) Convert the NNgen dataflow to a hardware description (Verilog HDL and IP-XACT)
    # --------------------

    # to Veriloggen object
    # targ = ng.to_veriloggen(outs, 'yolov3tiny', silent=silent,
    #                        config={'maxi_datawidth': axi_datawidth})

    # to IP-XACT (the method returns Veriloggen object, as well as to_veriloggen)
    targ = ng.to_ipxact(outs, 'yolov3tiny', silent=silent,
                        config={'maxi_datawidth': axi_datawidth})

    # to Verilog HDL RTL (the method returns a source code text)
    # rtl = ng.to_verilog(outs, 'yolov3tiny', silent=silent,
    #                    config={'maxi_datawidth': axi_datawidth})

    # --------------------
    # (6) Save the quantized weights
    # --------------------
    param_data = ng.export_ndarray(outs, chunk_size)
    param_bytes = len(param_data)
    np.savez_compressed(weight_filename, param_data)

    # --------------------
    # (7) Simulate the generated hardware by Veriloggen and Verilog simulator
    # --------------------

    if simtype is None:
        sys.exit()

    variable_addr = int(math.ceil((act.addr + act.memory_size) / chunk_size)) * chunk_size
    check0_addr = int(math.ceil((variable_addr + param_bytes) / chunk_size)) * chunk_size
    check1_addr = int(math.ceil((check0_addr + outs[0].memory_size) / chunk_size)) * chunk_size
    tmp_addr = int(math.ceil((check1_addr + outs[1].memory_size) / chunk_size)) * chunk_size

    memimg_datawidth = 32
    # mem = np.zeros([1024 * 1024 * 256 // (memimg_datawidth // 8)], dtype=np.int64)
    mem = np.zeros([1024 * 1024 * 256 // (memimg_datawidth // 8)], dtype=np.int16)
    mem = mem + [100]

    # placeholder
    axi.set_memory(mem, vact, memimg_datawidth,
                   act_dtype.width, act.addr,
                   max(int(math.ceil(axi_datawidth / act_dtype.width)), conv2d_par_ich))

    # parameters (variable and constant)
    axi.set_memory(mem, param_data, memimg_datawidth,
                   8, variable_addr)

    # verification data
    axi.set_memory(mem, vouts[0], memimg_datawidth,
                   act_dtype.width, check0_addr,
                   max(int(math.ceil(axi_datawidth / act_dtype.width)), conv2d_par_och))
    axi.set_memory(mem, vouts[1], memimg_datawidth,
                   act_dtype.width, check1_addr,
                   max(int(math.ceil(axi_datawidth / act_dtype.width)), conv2d_par_och))

    # test controller
    m = Module('test')
    params = m.copy_params(targ)
    ports = m.copy_sim_ports(targ)
    clk = ports['CLK']
    resetn = ports['RESETN']
    rst = m.Wire('RST')
    rst.assign(Not(resetn))

    # AXI memory model
    if sim_filename is None:
        sim_filename = os.path.splitext(os.path.basename(__file__))[0] + '.out'

    memimg_name = 'memimg_' + sim_filename

    memory = axi.AxiMemoryModel(m, 'memory', clk, rst,
                                datawidth=axi_datawidth,
                                memimg=mem, memimg_name=memimg_name,
                                memimg_datawidth=memimg_datawidth)
    memory.connect(ports, 'maxi')

    # AXI-Slave controller
    _saxi = vthread.AXIMLite(m, '_saxi', clk, rst, noio=True)
    _saxi.connect(ports, 'saxi')

    # timer
    time_counter = m.Reg('time_counter', 32, initval=0)
    seq = Seq(m, 'seq', clk, rst)
    seq(
        time_counter.inc()
    )

    def ctrl():
        for i in range(100):
            pass

        ng.sim.set_global_addrs(_saxi, tmp_addr)

        start_time = time_counter.value
        ng.sim.start(_saxi)

        print('# start')

        ng.sim.wait(_saxi)
        end_time = time_counter.value

        print('# end')
        print('# execution cycles: %d' % (end_time - start_time))

        # verify
        ok = True
        for bat in range(outs[0].shape[0]):
            for x in range(outs[0].shape[1]):
                orig = memory.read_word(bat * outs[0].aligned_shape[1] + x,
                                        outs[0].addr, act_dtype.width)
                check = memory.read_word(bat * outs[0].aligned_shape[1] + x,
                                         check0_addr, act_dtype.width)

                if vthread.verilog.NotEql(orig, check):
                    print('NG (', bat, x,
                          ') orig: ', orig, ' check: ', check)
                    ok = False
                # else:
                #    print('OK (', bat, x,
                #          ') orig: ', orig, ' check: ', check)

        for bat in range(outs[1].shape[0]):
            for x in range(outs[1].shape[1]):
                orig = memory.read_word(bat * outs[1].aligned_shape[1] + x,
                                        outs[1].addr, act_dtype.width)
                check = memory.read_word(bat * outs[1].aligned_shape[1] + x,
                                         check1_addr, act_dtype.width)

                if vthread.verilog.NotEql(orig, check):
                    print('NG (', bat, x,
                          ') orig: ', orig, ' check: ', check)
                    ok = False
                # else:
                #    print('OK (', bat, x,
                #          ') orig: ', orig, ' check: ', check)

        if ok:
            print('# verify: PASSED')
        else:
            print('# verify: FAILED')

        vthread.finish()

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    uut = m.Instance(targ, 'uut',
                     params=m.connect_params(targ),
                     ports=m.connect_ports(targ))

    # simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, resetn, m.make_reset(), period=100, polarity='low')

    init.add(
        Delay(10000000),
        Systask('finish'),
    )

    # output source code
    if verilog_filename is not None:
        m.to_verilog(verilog_filename)

    # run simulation
    sim = simulation.Simulator(m, sim=simtype)
    rslt = sim.run(outputfile=sim_filename)

    return rslt


if __name__ == '__main__':
    rslt = run(silent=False, verilog_filename='tmp.v')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import functools

import nngen.operator as operator

from . import util
from . import basic
from . import batchnormalization
from . import conv
from . import gemm


def _act_func(method, visitor, node):

    node_name = util.get_name(node)

    src_name = node.input[0]
    src_node = util.search_node_from_model(visitor.model, src_name)

    if (not visitor.disable_fusion and
        src_node is not None and
        src_node.op_type == 'BatchNormalization' and
            len(visitor.consumers[src_name]) == 1):

        src_op = batchnormalization.BatchNormalization(visitor, src_node,
                                                       act_func=method)
        visitor.operators[node_name] = src_op
        return src_op

    if (not visitor.disable_fusion and
        src_node is not None and
            src_node.op_type == 'Conv' and len(visitor.consumers[src_name]) == 1):

        src_op = conv.Conv(visitor, src_node, act_func=method)
        visitor.operators[node_name] = src_op
        return src_op

    if (not visitor.disable_fusion and
        src_node is not None and
            src_node.op_type == 'Gemm' and len(visitor.consumers[src_name]) == 1):

        src_op = gemm.Gemm(visitor, src_node, act_func=method)
        visitor.operators[node_name] = src_op
        return src_op

    return basic._elementwise(method, visitor, node)


def Relu(visitor, node):
    return _act_func(operator.relu, visitor, node)


def LeakyRelu(visitor, node):
    alpha = 0.01
    for attribute in node.attribute:
        if attribute.name == 'alpha':
            alpha = attribute.f

    rshift = 31
    slope = round(alpha * (2 ** 31))
    op = operator.get_leaky_relu_op(slope, rshift)
    return _act_func(op, visitor, node)


def Sigmoid(visitor, node):
    return _act_func(operator.sigmoid, visitor, node)



from __future__ import absolute_import
from __future__ import print_function

import os
import sys

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng
import veriloggen

import onnx_vgg11


act_dtype = ng.int8
weight_dtype = ng.int8
bias_dtype = ng.int32
scale_dtype = ng.int8
with_batchnorm = True
disable_fusion = False
conv2d_par_ich = 1
conv2d_par_och = 1
conv2d_par_col = 1
conv2d_par_row = 1
conv2d_concur_och = None
conv2d_stationary = 'filter'
pool_par = 1
elem_par = 1
chunk_size = 64
axi_datawidth = 32


def test(request, silent=True):
    veriloggen.reset()

    simtype = request.config.getoption('--sim')

    rslt = onnx_vgg11.run(act_dtype, weight_dtype,
                          bias_dtype, scale_dtype,
                          with_batchnorm, disable_fusion,
                          conv2d_par_ich, conv2d_par_och, conv2d_par_col, conv2d_par_row,
                          conv2d_concur_och, conv2d_stationary,
                          pool_par, elem_par,
                          chunk_size,
                          axi_datawidth, silent,
                          filename=None, simtype=simtype,
                          outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')

    verify_rslt = [line for line in rslt.splitlines() if line.startswith('# verify:')][0]
    assert(verify_rslt == '# verify: PASSED')


if __name__ == '__main__':
    rslt = onnx_vgg11.run(act_dtype, weight_dtype,
                          bias_dtype, scale_dtype,
                          with_batchnorm, disable_fusion,
                          conv2d_par_ich, conv2d_par_och, conv2d_par_col, conv2d_par_row,
                          conv2d_concur_och, conv2d_stationary,
                          pool_par, elem_par,
                          chunk_size,
                          axi_datawidth, silent=False,
                          filename='tmp.v',
                          outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import collections

import nngen.storage as storage
import nngen.operator as operator
import nngen.dtype_list as dtype_list

from . import util


def Conv(visitor, node,
         batchnorm_scale=None, batchnorm_bias=None, act_func=None):

    # input, filter
    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    input = srcs[0]
    filter = srcs[1]

    # transpose data layout to nngen-compatible format
    input = util.transpose_layout(input, visitor.nngen_input_layout, visitor.onnx_input_layout)
    filter = util.transpose_layout(filter, visitor.nngen_filter_layout, visitor.onnx_filter_layout)

    bias = srcs[2] if len(srcs) > 2 else None

    name = util.get_name(node)

    scale_name = '_'.join(['onnx', name, 'conv.scale'])
    scale_dtype = visitor.default_scale_dtype
    scale_shape = batchnorm_scale.shape if batchnorm_scale is not None else (1,)
    scale = storage.variable(dtype=scale_dtype, shape=scale_shape, name=scale_name)
    scale_value = batchnorm_scale if batchnorm_scale is not None else [1]
    scale.set_value(scale_value)
    visitor.variables[scale_name] = scale

    if bias is None and batchnorm_bias is not None:
        bias_name = '_'.join(['onnx', name, 'conv.bias'])
        bias_dtype = visitor.default_bias_dtype
        bias_shape = batchnorm_bias.shape
        bias = storage.variable(dtype=bias_dtype, shape=bias_shape, name=bias_name)
        bias_value = batchnorm_bias / batchnorm_scale
        bias.set_value(bias_value)
        visitor.variables[bias_name] = bias

    elif bias is not None and batchnorm_bias is not None:
        bias.dtype = visitor.default_bias_dtype
        bias_value = batchnorm_bias / batchnorm_scale + bias.value
        bias.set_value(bias_value)

    elif bias is not None:
        bias.dtype = visitor.default_bias_dtype

    #rshift_out_name = '_'.join(['onnx', name, 'conv.rshift_out'])
    #rshift_out_width = filter.dtype.width
    #rshift_out_dtype = dtype_list.dtype_int(rshift_out_width, signed=False)
    #rshift_out_shape = (1,)
    #rshift_out = storage.variable(dtype=scale_dtype, shape=scale_shape, name=rshift_out_name)
    #visitor.variables[rshift_out_name] = rshift_out
    rshift_out = 0

    if name in visitor.value_dtypes:
        dtype = visitor.value_dtypes[name]
    else:
        dtype = visitor.default_operator_dtype

    if dtype.width >= 16:
        sum_dtype = dtype_list.dtype_int(dtype.width * 4)
    else:
        sum_dtype = dtype_list.int32

    strides = [1, 1, 1, 1]  # B, H, W, C
    padding = [0, 0, 0, 0]  # Top, Bottom, Left, Right

    for attribute in node.attribute:
        if attribute.name == 'auto_pad':
            padding = 'SAME'

        elif attribute.name == 'pads':
            padding[0] = attribute.ints[0]
            padding[1] = attribute.ints[1]
            padding[2] = attribute.ints[2]
            padding[3] = attribute.ints[3]
            padding = tuple(padding)

        elif attribute.name == 'strides':
            strides[1] = attribute.ints[0]
            strides[2] = attribute.ints[1]
            strides = tuple(strides)

    args = [input, filter]

    kwargs = collections.OrderedDict()
    kwargs['strides'] = strides
    kwargs['bias'] = bias
    kwargs['scale'] = scale
    kwargs['rshift_out'] = rshift_out
    kwargs['act_func'] = act_func
    kwargs['padding'] = padding
    kwargs['dtype'] = dtype
    kwargs['sum_dtype'] = sum_dtype
    kwargs['name'] = name

    c = operator.conv2d(*args, **kwargs)
    c.layout = visitor.nngen_input_layout
    c.onnx_layout = visitor.onnx_input_layout

    return c



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np
import functools
import collections

import nngen.operator as operator

from . import util


def Reshape(visitor, node, no_transpose=False):

    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    srcs = [util.optimize_to_raw_value(src) for src in srcs]

    input = srcs[0]
    out_shape = srcs[1]

    if not isinstance(out_shape, (tuple, list, np.ndarray)):
        raise TypeError('shape must be tuple, list, or np.ndarray, not %s' %
                        str(type(out_shape)))

    if isinstance(out_shape, np.ndarray):
        out_shape = out_shape.tolist()

    if isinstance(input, (tuple, list)):
        input = np.array(input)

    if isinstance(input, np.ndarray):
        c = np.reshape(input, out_shape)
        return c

    length = input.get_length()
    out_shape = util.infer_shape(out_shape, length)

    out_layout = None
    out_onnx_layout = None

    if input.get_onnx_layout() is not None and input.get_onnx_layout() is not None:
        layout = input.get_layout()
        onnx_layout = input.get_onnx_layout()
        onnx_shape = [input.shape[layout.index(ol)]
                      for ol in onnx_layout]
        out_onnx_layout = make_layout(onnx_shape, out_shape, onnx_layout)

        if no_transpose or input.get_layout() == input.get_onnx_layout():
            out_layout = out_onnx_layout

        elif is_split_reshape(onnx_shape, out_shape, onnx_layout, out_onnx_layout):
            # keep the original input layout as possible
            layout_map = layout_matching(onnx_shape, out_shape, onnx_layout, out_onnx_layout)

            out_layout = []
            for l in layout:
                key = (l,)
                out_layout.extend(layout_map[key])

            out_layout = tuple(out_layout)
            out_shape = [out_shape[out_onnx_layout.index(ol)]
                         for ol in out_layout]

        else:
            # transpose the input tensor for applying Reshape to the correct layout
            out_layout = out_onnx_layout

            perm = [input.get_layout().index(l) for l in input.get_onnx_layout()]
            onnx_perm = [i for i, l in enumerate(input.get_onnx_layout())]

            input = operator.transpose(input, perm)
            input.implicit = True
            input.transpose_onnx_perm = onnx_perm

            input.layout = input.get_onnx_layout()

    name = util.get_name(node)

    c = operator.reshape(input, out_shape, name=name)
    c.layout = out_layout
    c.onnx_layout = out_onnx_layout

    return c


def make_layout(src_shape, dst_shape, src_layout, prefix='X'):

    si = 0
    di = 0
    tmp = 0
    dst_layout = []

    while True:
        src_size = functools.reduce(lambda x, y: x * y, src_shape[:si], 1)
        dst_size = functools.reduce(lambda x, y: x * y, dst_shape[:di], 1)

        if len(dst_layout) == len(dst_shape):
            break

        if si >= len(src_shape) and di >= len(dst_shape):
            break

        if src_size == dst_size:
            ss = src_shape[si] if si < len(src_shape) else 1
            ds = dst_shape[di] if di < len(dst_shape) else 1
            if ss == ds:
                if si < len(src_shape):
                    dst_layout.append(src_layout[si])
                    si += 1
                else:
                    dst_layout.append('%s%d' % (prefix, tmp))
                    tmp += 1

                di += 1
            else:
                si += 1

        elif src_size > dst_size:
            dst_layout.append('%s%d' % (prefix, tmp))
            tmp += 1
            di += 1

        elif src_size < dst_size:
            si += 1

    return tuple(dst_layout)


def layout_matching(src_shape, dst_shape, src_layout, dst_layout):

    si = 0
    di = 0
    key = []
    value = []
    matching = collections.OrderedDict()

    while True:
        src_size = functools.reduce(lambda x, y: x * y, src_shape[:si], 1)
        dst_size = functools.reduce(lambda x, y: x * y, dst_shape[:di], 1)

        if si >= len(src_shape) and di >= len(dst_shape):
            if len(key) > 0 and len(value) > 0:
                matching[tuple(key)] = tuple(value)
            break

        if src_size == dst_size:
            if len(key) > 0 and len(value) > 0:
                matching[tuple(key)] = tuple(value)
                key = []
                value = []

            key.append(src_layout[si])
            si += 1
            while si < len(src_shape) and src_shape[si] == 1:
                key.append(src_layout[si])
                si += 1

            value.append(dst_layout[di])
            di += 1
            while di < len(dst_shape) and dst_shape[di] == 1:
                value.append(dst_layout[di])
                di += 1

        elif src_size > dst_size:
            value.append(dst_layout[di])
            di += 1

        elif src_size < dst_size:
            key.append(src_layout[si])
            si += 1

    return matching


def is_split_reshape(src_shape, dst_shape, src_layout, dst_layout):
    matching = layout_matching(src_shape, dst_shape, src_layout, dst_layout)

    for sl in src_layout:
        key = (sl,)
        if key not in matching:
            return False

    return True



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division


def Shape(visitor, node):

    input = visitor.visit(node.input[0])

    shape = input.shape
    if (input.get_layout() is not None and input.get_onnx_layout() is not None and
            input.get_layout() != input.get_onnx_layout()):
        shape = [shape[input.get_layout().index(l)] for l in input.get_onnx_layout()]

    return tuple(shape)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np

import nngen.operator as operator

from . import util


def Concat(visitor, node):

    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    axis = None
    for attribute in node.attribute:
        if attribute.name == 'axis':
            axis = attribute.i

    srcs = [util.optimize_to_raw_value(src) for src in srcs]

    all_ndarray = True
    for src in srcs:
        if not isinstance(src, np.ndarray):
            all_ndarray = False
            break

    if all_ndarray:
        return np.concatenate(srcs, axis)

    name = util.get_name(node)

    scales = [1.0 for src in srcs]
    shamt = 0

    layout = None
    onnx_layout = None
    for src in srcs:
        l = src.get_layout()
        if l is None:
            continue
        if layout is None:
            layout = l
        elif layout != l:
            raise ValueError("layout mismatch: '%s' != '%s'" % (layout, l))

        l = src.get_onnx_layout()
        if l is None:
            continue
        if onnx_layout is None:
            onnx_layout = l
        elif onnx_layout != l:
            raise ValueError("onnx_layout mismatch: '%s' != '%s'" % (onnx_layout, l))

    if layout is not None and onnx_layout is not None:
        axis = layout.index(onnx_layout[axis])

    c = operator.scaled_concat(srcs, scales, shamt, axis, name=name)

    return c



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import math
from collections import OrderedDict

import veriloggen as vg

import nngen.basic_types as bt
import nngen.util as util

from .pool import _pool


class _pool_serial(_pool):
    # shape order
    # - value:  'NHWC' (batch, height, width, channel)
    # - ksize:  'NHWC' (batch, height, width, channel)

    input_chainable = False
    output_chainable = False

    control_param_custom_width = {'act_offset_values': bt.get_maxi_addrwidth}
    control_param_custom_signed = {'act_offset_values': True}

    def __sub_str__(self):
        no_reuse = ' no_reuse' if self.no_reuse else ''
        return ''.join([_pool.__sub_str__(self), no_reuse])

    def __init__(self, value, ksize, strides, padding='SAME',
                 dtype=None, name=None, par=1,
                 value_ram_size=None, out_ram_size=None):

        self.no_reuse = True
        for k, s in zip(ksize, strides):
            if k > s:
                self.no_reuse = False
                raise ValueError("_pool_serial does not support"
                                 " 'ksize > stride': %s > %s" %
                                 (str(ksize), str(strides)))

        _pool.__init__(self, value, ksize, strides, padding,
                       dtype, name, par,
                       value_ram_size, out_ram_size)

    def get_required_rams(self):
        ksize_ch = self.ksize[-1]
        ksize_col = self.ksize[-2]
        ksize_row = self.ksize[-3]
        ksize_bat = self.ksize[-4]

        act = self.args[0]
        act_shape = act.get_aligned_shape()
        act_num_ch = act_shape[-1]
        act_num_col = act_shape[-2]
        act_num_row = act_shape[-3]
        act_num_bat = act_shape[-4]

        out_shape = self.get_aligned_shape()
        out_num_ch = out_shape[-1]
        out_num_col = out_shape[-2]
        out_num_row = out_shape[-3]
        out_num_bat = out_shape[-4]

        input_min_size = (int(math.ceil(act_num_ch / self.par))
                          * act_num_col * ksize_row * 2)
        if self.value_ram_size is not None and input_min_size < self.value_ram_size:
            input_min_size = self.value_ram_size
        input_width = act.get_ram_width() * self.par

        output_min_size = (int(math.ceil(out_num_ch / self.par))
                           * out_num_col * 2)
        if self.out_ram_size is not None and output_min_size < self.out_ram_size:
            output_min_size = self.out_ram_size
        output_width = self.get_ram_width() * self.par

        inputs = []
        inputs.append((input_width, input_min_size))

        outputs = []
        outputs.append((output_width, output_min_size))

        temps = []

        return inputs, outputs, temps

    def get_stream_hash(self):
        base = bt._Operator.get_stream_hash(self)
        ksize_col = self.ksize[-2]
        ksize_row = self.ksize[-3]
        return (base, ksize_col, ksize_row, self.no_reuse, self.par)

    def get_stream_func(self):

        def func(strm):
            ksize_col = self.ksize[-2]
            ksize_row = self.ksize[-3]

            # iteration size
            size = strm.parameter(datawidth=(ksize_col * ksize_row).bit_length(),
                                  signed=False)

            # vec_act
            arg = self.args[0]
            datawidth = arg.get_op_width()
            vec_datawidth = datawidth * self.par
            point = arg.get_op_point()
            signed = arg.get_signed()

            vec_act_var = strm.source(datawidth=vec_datawidth, signed=False)

            # mask
            mask_vector = strm.parameter(datawidth=ksize_col * ksize_row, signed=False)
            mask_counter = strm.Counter(size=size)
            mask = strm.Pointer(mask_vector, mask_counter)

            # vec_act -> act
            act_var_list = []
            split_vec_act_var = strm.Split(vec_act_var,
                                           datawidth, point, signed, reverse=True)

            for i in range(self.par):
                if self.par == 1:
                    act_var = strm.ReinterpretCast(vec_act_var, datawidth, point, signed)
                else:
                    act_var = split_vec_act_var[i]

                act_var_list.append(act_var)

            pad_value = self.get_pad_value(strm)

            # vector parallel processing
            out_vars = []
            out_valids = []

            for i, act_var in enumerate(act_var_list):
                masked_var = strm.Mux(mask, pad_value, act_var)
                out_var, out_valid = self.pool_op(strm, i, masked_var, size)

                width = self.get_op_width()
                point = self.get_op_point()
                signed = self.get_signed()

                out_var = bt.out_rcast(strm, out_var, width, point, signed)
                out_vars.append(out_var)
                out_valids.append(out_valid)

            if self.par == 1:
                vec_out_var = out_vars[0]
            else:
                vec_out_var = strm.Cat(*reversed(out_vars))

            vec_out_valid = out_valids[0]
            strm.sink(vec_out_var, when=vec_out_valid)

        return func

    def pool_op(self, strm, index, var, size):
        # return value
        raise NotImplementedError('not implemented')

    def get_control_param_values(self):
        act = self.args[0]

        ksize_ch = self.ksize[-1]
        ksize_col = self.ksize[-2]
        ksize_row = self.ksize[-3]
        ksize_bat = self.ksize[-4]

        act_shape = act.get_aligned_shape()
        act_num_ch = act_shape[-1]
        act_num_col = act_shape[-2]
        act_num_row = act_shape[-3]
        act_num_bat = act_shape[-4]

        # stride_ch = self.strides[-1]  # always 1
        stride_col = self.strides[-2]  # width
        stride_row = self.strides[-3]  # height
        stride_bat = self.strides[-4]  # always 1

        out_shape = self.get_aligned_shape()
        out_num_ch = out_shape[-1]
        out_num_col = out_shape[-2]
        out_num_row = out_shape[-3]
        out_num_bat = out_shape[-4]

        if isinstance(self.padding, str) and self.padding == 'SAME':
            pad_col, pad_col_left, pad_col_right = util.pad_size_split(
                act_num_col, ksize_col, stride_col)
            pad_row, pad_row_top, pad_row_bottom = util.pad_size_split(
                act_num_row, ksize_row, stride_row)
        elif isinstance(self.padding, int):
            pad_col = self.padding * 2
            pad_col_left = self.padding
            pad_col_right = self.padding
            pad_row = self.padding * 2
            pad_row_top = self.padding
            pad_row_bottom = self.padding
        elif isinstance(self.padding, (tuple, list)):
            pad_col = self.padding[2] + self.padding[3]
            pad_col_left = self.padding[2]
            pad_col_right = self.padding[3]
            pad_row = self.padding[0] + self.padding[1]
            pad_row_top = self.padding[0]
            pad_row_bottom = self.padding[1]
        else:
            pad_col = 0
            pad_col_left = 0
            pad_col_right = 0
            pad_row = 0
            pad_row_top = 0
            pad_row_bottom = 0

        # for __str__
        self.pad_col_left_value = pad_col_left
        self.pad_col_right_value = pad_col_right
        self.pad_row_top_value = pad_row_top
        self.pad_row_bottom_value = pad_row_bottom

        max_col_count = act_num_col + pad_col + 1 - ksize_col - stride_col
        if max_col_count < 0:
            max_col_count = 0

        max_row_count = act_num_row + pad_row + 1 - ksize_row - stride_row
        if max_row_count < 0:
            max_row_count = 0

        max_bat_count = act_num_bat - stride_bat
        if max_bat_count < 0:
            max_bat_count = 0

        aligned_act_num_ch = bt.align_word(act_num_ch,
                                           act.get_word_alignment())

        act_step = bt.to_byte(aligned_act_num_ch * act.get_ram_width())

        act_offset_values = []
        for y in range(ksize_row):
            v = act_num_col * (y - pad_row_top) * act_step
            act_offset_values.append(v)

        act_row_step = act_step * act_num_col * stride_row
        act_bat_step = act_step * act_num_col * act_num_row

        act_read_size = (int(math.ceil(aligned_act_num_ch / self.par))
                         * act_num_col)
        act_read_block = int(math.ceil(aligned_act_num_ch / self.par))

        out_step = bt.to_byte(bt.align_word(out_num_ch, self.get_word_alignment())
                              * self.get_ram_width())

        out_row_step = out_step * out_num_col
        out_bat_step = out_step * out_num_col * out_num_row

        out_write_size = (int(math.ceil(out_num_ch / self.par))
                          * out_num_col)

        stream_size = int(math.ceil(aligned_act_num_ch / self.par))

        if pad_col_left == 0:
            col_select_initval = 0
        else:
            col_select_initval = (ksize_col - pad_col_left) % ksize_col

        stride_col_mod_ksize = stride_col % ksize_col
        ksize_col_minus_stride_col_mod = ksize_col - stride_col_mod_ksize

        local_pad_offset = -1 * pad_col_left * stream_size

        inc_act_laddr = ksize_col * act_read_block

        inc_out_laddr = int(math.ceil(out_num_ch / self.par))

        return OrderedDict([('act_num_col', act_num_col),
                            ('act_num_row', act_num_row),
                            ('stride_col', stride_col),
                            ('stride_row', stride_row),
                            ('out_num_col', out_num_col),
                            ('out_num_row', out_num_row),
                            ('pad_col_left', pad_col_left),
                            ('pad_row_top', pad_row_top),
                            ('max_col_count', max_col_count),
                            ('max_row_count', max_row_count),
                            ('max_bat_count', max_bat_count),
                            ('act_offset_values', act_offset_values),
                            ('act_row_step', act_row_step),
                            ('act_bat_step', act_bat_step),
                            ('act_read_size', act_read_size),
                            ('act_read_block', act_read_block),
                            ('out_row_step', out_row_step),
                            ('out_bat_step', out_bat_step),
                            ('out_write_size', out_write_size),
                            ('stream_size', stream_size),
                            ('col_select_initval', col_select_initval),
                            ('stride_col_mod_ksize', stride_col_mod_ksize),
                            ('ksize_col_minus_stride_col_mod', ksize_col_minus_stride_col_mod),
                            ('local_pad_offset', local_pad_offset),
                            ('inc_act_laddr', inc_act_laddr),
                            ('inc_out_laddr', inc_out_laddr)])

    def control_sequence(self, fsm):
        ksize_ch = self.ksize[-1]
        ksize_col = self.ksize[-2]
        ksize_row = self.ksize[-3]
        ksize_bat = self.ksize[-4]

        self.stride_bat = 1

        act_ram = self.input_rams[0]
        out_ram = self.output_rams[0]

        act_base_offset = self.m.Wire(self._name('act_base_offset'),
                                      self.maxi.addrwidth, signed=True)
        act_base_offset_row = self.m.Reg(self._name('act_base_offset_row'),
                                         self.maxi.addrwidth, initval=0, signed=True)
        act_base_offset_bat = self.m.Reg(self._name('act_base_offset_bat'),
                                         self.maxi.addrwidth, initval=0, signed=True)

        act_base_offset.assign(act_base_offset_row
                               + act_base_offset_bat)

        out_base_offset = self.m.Wire(self._name('out_base_offset'),
                                      self.maxi.addrwidth, signed=True)
        out_base_offset_row = self.m.Reg(self._name('out_base_offset_row'),
                                         self.maxi.addrwidth, initval=0, signed=True)
        out_base_offset_bat = self.m.Reg(self._name('out_base_offset_bat'),
                                         self.maxi.addrwidth, initval=0, signed=True)

        out_base_offset.assign(out_base_offset_row
                               + out_base_offset_bat)

        col_count = self.m.Reg(self._name('col_count'),
                               self.maxi.addrwidth, initval=0)
        row_count = self.m.Reg(self._name('row_count'),
                               self.maxi.addrwidth, initval=0)
        bat_count = self.m.Reg(self._name('bat_count'),
                               self.maxi.addrwidth, initval=0)

        if not self.no_reuse:
            col_select = self.m.Reg(self._name('col_select'),
                                    bt.log_width(ksize_col),
                                    initval=0)
            row_select = self.m.Reg(self._name('row_select'),
                                    bt.log_width(ksize_row),
                                    initval=0)

        prev_row_count = self.m.Reg(self._name('prev_row_count'),
                                    self.maxi.addrwidth, initval=0)
        prev_bat_count = self.m.Reg(self._name('prev_bat_count'),
                                    self.maxi.addrwidth, initval=0)

        if not self.no_reuse:
            prev_row_select = self.m.Reg(self._name('prev_row_select'),
                                         bt.log_width(ksize_row),
                                         initval=0)

        stream_act_local = self.m.Reg(self._name('stream_act_local'),
                                      self.maxi.addrwidth, initval=0)
        stream_out_local = self.m.Reg(self._name('stream_out_local'),
                                      self.maxi.addrwidth, initval=0)

        # double buffer control
        act_page = self.m.Reg(self._name('act_page'), initval=0)
        act_page_comp_offset = self.m.Reg(self._name('act_page_comp_offset'),
                                          self.maxi.addrwidth, initval=0)
        act_page_dma_offset = self.m.Reg(self._name('act_page_dma_offset'),
                                         self.maxi.addrwidth, initval=0)

        out_page = self.m.Reg(self._name('out_page'), initval=0)
        out_page_comp_offset = self.m.Reg(self._name('out_page_comp_offset'),
                                          self.maxi.addrwidth, initval=0)
        out_page_dma_offset = self.m.Reg(self._name('out_page_dma_offset'),
                                         self.maxi.addrwidth, initval=0)

        act_page_size = act_ram.length // 2
        out_page_size = out_ram.length // 2

        skip_read_act = self.m.Reg(self._name('skip_read_act'), initval=0)
        skip_comp = self.m.Reg(self._name('skip_comp'), initval=0)
        skip_write_out = self.m.Reg(self._name('skip_write_out'), initval=0)

        comp_count = self.m.Reg(self._name('comp_count'),
                                self.maxi.addrwidth, initval=0)
        out_count = self.m.Reg(self._name('out_count'),
                               self.maxi.addrwidth, initval=0)

        # --------------------
        # initialization phase
        # --------------------
        # ReadAct: offset
        fsm(
            act_base_offset_row(0),
            act_base_offset_bat(0)
        )

        act_offsets = []
        for v in self.act_offset_values:
            act_offset = act_base_offset + v
            act_offsets.append(act_offset)

        # ReadAct: DMA flag
        dma_pad_masks = []

        for y in range(ksize_row):
            v = vg.Ors((row_count + y < self.pad_row_top),
                       (row_count + y >= self.act_num_row + self.pad_row_top))
            dma_pad_mask = self.m.Wire(
                self._name('dma_pad_mask_%d' % y))
            dma_pad_mask.assign(v)
            dma_pad_masks.append(dma_pad_mask)

        # ReadAct: double buffer control
        fsm(
            act_page(0),
            act_page_comp_offset(0),
            act_page_dma_offset(0)
        )

        # WriteOutput: offset
        fsm(
            out_base_offset_row(0),
            out_base_offset_bat(0)
        )

        out_offset = out_base_offset

        # WriteOut: double buffer control
        fsm(
            out_page(0),
            out_page_comp_offset(0),
            out_page_dma_offset(0)
        )

        # counter
        fsm(
            row_count(0),
            bat_count(0),
            prev_row_count(0),
            prev_bat_count(0)
        )

        if not self.no_reuse:
            fsm(
                row_select(0),
                prev_row_select(0)
            )

        # double buffer control
        fsm(
            skip_read_act(0),
            skip_comp(0),
            skip_write_out(1)
        )

        fsm(
            out_count(0)
        )

        state_init = fsm.current

        fsm.goto_next()

        # --------------------
        # ReadAct phase
        # --------------------
        state_read_act = fsm.current
        fsm.goto_next()

        act_gaddrs = []
        for act_offset in act_offsets:
            act_gaddr = self.arg_objaddrs[0] + act_offset
            act_gaddrs.append(act_gaddr)

        if not self.no_reuse:
            mux_act_gaddr_values = mux_1d(act_gaddrs, row_select, ksize_row)
            mux_act_gaddrs = []
            for i, mux_act_gaddr_value in enumerate(mux_act_gaddr_values):
                mux_act_gaddr = self.m.Wire(self._name('mux_act_gaddr_%d' % i),
                                            self.maxi.addrwidth)
                mux_act_gaddr.assign(mux_act_gaddr_value)
                mux_act_gaddrs.append(mux_act_gaddr)

            mux_dma_pad_mask_values = mux_1d(
                dma_pad_masks, row_select, ksize_row)
            mux_dma_pad_masks = []
            for i, mux_dma_pad_mask_value in enumerate(mux_dma_pad_mask_values):
                mux_dma_pad_mask = self.m.Wire(
                    self._name('mux_dma_pad_mask_%d' % i))
                mux_dma_pad_mask.assign(mux_dma_pad_mask_value)
                mux_dma_pad_masks.append(mux_dma_pad_mask)

        else:
            mux_act_gaddrs = act_gaddrs
            mux_dma_pad_masks = dma_pad_masks

        bt.bus_lock(self.maxi, fsm)

        act_laddr = act_page_dma_offset

        for (act_gaddr, dma_pad_mask) in zip(
                mux_act_gaddrs, mux_dma_pad_masks):
            begin_state_read = fsm.current
            fsm.goto_next()

            bt.dma_read(self.maxi, fsm, act_ram, act_laddr,
                        act_gaddr, self.act_read_size, port=1)

            end_state_read = fsm.current

            fsm.If(dma_pad_mask).goto_from(
                begin_state_read, end_state_read)

            act_laddr += self.act_read_size

        bt.bus_unlock(self.maxi, fsm)

        fsm.goto_next()
        state_read_act_end = fsm.current
        fsm.If(skip_read_act).goto_from(state_read_act, state_read_act_end)

        # --------------------
        # Comp phase
        # --------------------
        state_comp = fsm.current

        # waiting for previous DMA write completion
        bt.dma_wait_write_idle(self.maxi, fsm)

        state_comp_after_sync = fsm.current

        # Stream Control FSM
        comp_fsm = vg.FSM(self.m, self._name('comp_fsm'), self.clk, self.rst)
        comp_state_init = comp_fsm.current
        comp_fsm.If(fsm.state == state_comp_after_sync, vg.Not(skip_comp)).goto_next()

        # when comp_fsm is available, go to the next phase
        fsm.If(comp_fsm.state == comp_state_init).goto_next()

        # local address
        comp_fsm(
            stream_act_local(self.local_pad_offset)
        )

        comp_fsm(
            stream_out_local(0)
        )

        # count and sel
        comp_fsm(
            col_count(0)
        )

        if not self.no_reuse:
            comp_fsm(
                col_select(self.col_select_initval)
            )

        act_page_comp_offset_buf = self.m.Reg(self._name('act_page_comp_offset_buf'),
                                              self.maxi.addrwidth, initval=0)
        out_page_comp_offset_buf = self.m.Reg(self._name('out_page_comp_offset_buf'),
                                              self.maxi.addrwidth, initval=0)
        row_count_buf = self.m.Reg(self._name('row_count_buf'),
                                   self.maxi.addrwidth, initval=0)

        if not self.no_reuse:
            row_select_buf = self.m.Reg(self._name('row_select_buf'),
                                        bt.log_width(ksize_row),
                                        initval=0)
        comp_fsm(
            act_page_comp_offset_buf(act_page_comp_offset),
            out_page_comp_offset_buf(out_page_comp_offset),
            row_count_buf(row_count)
        )

        if not self.no_reuse:
            comp_fsm(
                row_select_buf(row_select)
            )

        comp_fsm.goto_next()

        # repeat comp
        comp_state_rep = comp_fsm.current

        # pad_mask
        stream_pad_masks = []

        for y in range(ksize_row):
            for x in range(ksize_col):
                stream_col_count = col_count + x
                stream_row_count = row_count_buf + y
                v = vg.Ors((stream_col_count < self.pad_col_left),
                           (stream_col_count >= self.act_num_col + self.pad_col_left),
                           (stream_row_count < self.pad_row_top),
                           (stream_row_count >= self.act_num_row + self.pad_row_top))
                stream_pad_mask = self.m.Wire(
                    self._name('stream_pad_mask_%d_%d' % (y, x)))
                stream_pad_mask.assign(v)
                stream_pad_masks.append(stream_pad_mask)

        if not self.no_reuse:
            stream_pad_mask_2d = line_to_2d(stream_pad_masks, ksize_col)
            stream_pad_mask_2d_mux = mux_2d(stream_pad_mask_2d,
                                            col_select, row_select_buf,
                                            ksize_col, ksize_row)
            stream_pad_masks = [flatten for inner in stream_pad_mask_2d_mux
                                for flatten in inner]

        stream_pad_masks_reg = self.m.Reg(self._name('stream_pad_masks'),
                                          len(stream_pad_masks), initval=0)
        comp_fsm(
            stream_pad_masks_reg(vg.Cat(*reversed(stream_pad_masks)))
        )
        comp_fsm.goto_next()

        # busy check
        self.stream.source_join(comp_fsm)

        stream_masks = stream_pad_masks_reg

        # set_parameter
        name = list(self.stream.parameters.keys())[0]
        self.stream.set_parameter(comp_fsm, name, ksize_col * ksize_row)
        comp_fsm.set_index(comp_fsm.current - 1)

        name = list(self.stream.parameters.keys())[1]
        self.stream.set_parameter(comp_fsm, name, stream_masks)
        comp_fsm.set_index(comp_fsm.current - 1)

        # set_source
        name = list(self.stream.sources.keys())[0]
        local = stream_act_local + act_page_comp_offset_buf
        pat = ((ksize_col, self.act_read_block),
               (ksize_row, self.act_read_size),
               (self.stream_size, 1))
        self.stream.set_source_pattern(comp_fsm, name, act_ram,
                                       local, pat)
        comp_fsm.set_index(comp_fsm.current - 1)

        # set_sink
        name = list(self.stream.sinks.keys())[0]
        local = stream_out_local + out_page_comp_offset_buf
        self.stream.set_sink(comp_fsm, name, out_ram, local, self.stream_size)

        # stream run (async)
        self.stream.run(comp_fsm)

        # stream_act_local
        comp_fsm(
            stream_act_local.add(self.inc_act_laddr)
        )
        comp_fsm.If(col_count >= self.max_col_count)(
            stream_act_local(self.local_pad_offset)
        )

        # stream_out_local
        comp_fsm(
            stream_out_local.add(self.inc_out_laddr)
        )
        comp_fsm.If(col_count >= self.max_col_count)(
            stream_out_local(0)
        )

        # counter
        comp_fsm(
            col_count.add(self.stride_col)
        )
        comp_fsm.If(col_count >= self.max_col_count)(
            col_count(0)
        )

        if not self.no_reuse:
            comp_fsm(
                col_select.add(self.stride_col_mod_ksize)
            )
            comp_fsm.If(col_select + self.stride_col_mod_ksize >= ksize_col)(
                col_select.sub(self.ksize_col_minus_stride_col_mod)
            )

            comp_fsm.If(col_count >= self.max_col_count)(
                col_select(self.col_select_initval)
            )

        # repeat
        comp_fsm.goto(comp_state_rep)
        comp_fsm.If(col_count >= self.max_col_count).goto_init()

        # sync with WriteOut control
        comp_fsm.seq.If(fsm.state == state_init)(
            comp_count(0)
        )
        comp_fsm.seq.If(self.stream.sink_stop)(
            comp_count.add(self.inc_out_laddr)
        )

        # --------------------
        # WriteOut phase
        # --------------------
        state_write_out = fsm.current
        fsm.goto_next()

        # sync with Comp control
        fsm.If(comp_count >= out_count + self.out_write_size).goto_next()

        out_laddr = out_page_dma_offset
        out_gaddr = self.objaddr + out_offset

        bt.bus_lock(self.maxi, fsm)

        bt.dma_write(self.maxi, fsm, out_ram, out_laddr,
                     out_gaddr, self.out_write_size, port=1, use_async=True)

        bt.bus_unlock(self.maxi, fsm)

        fsm(
            out_count.add(self.out_write_size)
        )

        fsm.goto_next()

        state_write_out_end = fsm.current
        fsm.If(skip_write_out).goto_from(state_write_out, state_write_out_end)

        # --------------------
        # update for next iteration
        # --------------------
        # ReadAct: offset
        fsm(
            act_base_offset_row.add(self.act_row_step)
        )
        fsm.If(row_count >= self.max_row_count)(
            act_base_offset_row(0),
            act_base_offset_bat.add(self.act_bat_step)
        )
        fsm.If(row_count >= self.max_row_count,
               bat_count >= self.max_bat_count)(
            act_base_offset_bat(0)
        )

        # ReadAct: counter
        fsm(
            row_count.add(self.stride_row)
        )
        fsm.If(row_count >= self.max_row_count)(
            row_count(0),
            bat_count.add(self.stride_bat)
        )
        fsm.If(row_count >= self.max_row_count,
               bat_count >= self.max_bat_count)(
            bat_count(0)
        )

        if not self.no_reuse:
            fsm.If(self.stride_row < ksize_row)(
                row_select.add(self.stride_row),
                prev_row_select(row_select)
            )
            fsm.If(self.stride_row < ksize_row,
                   row_select + self.stride_row >= ksize_row)(
                row_select(row_select - (vg.Int(ksize_row) - self.stride_row)),
                prev_row_select(row_select)
            )
            fsm.If(vg.Not(self.stride_row < ksize_row))(
                row_select(0),
                prev_row_select(0)
            )

            fsm.If(row_count >= self.max_row_count)(
                row_select(0),
                prev_row_select(0)
            )

        # ReadAct and Comp: double buffer
        fsm.If(vg.Not(act_page))(
            act_page_comp_offset(act_page_size),
            act_page_dma_offset(act_page_size),
            act_page(1)
        )
        fsm.If(act_page)(
            act_page_comp_offset(0),
            act_page_dma_offset(0),
            act_page(0)
        )

        # WriteOut: counter
        fsm.If(vg.Not(skip_write_out))(
            out_base_offset_row.add(self.out_row_step)
        )
        fsm.If(vg.Not(skip_write_out),
               prev_row_count >= self.max_row_count)(
            out_base_offset_row(0),
            out_base_offset_bat.add(self.out_bat_step)
        )
        fsm.If(vg.Not(skip_write_out),
               prev_row_count >= self.max_row_count,
               prev_bat_count >= self.max_bat_count)(
            out_base_offset_bat(0)
        )

        # WriteOut and Comp: double buffer
        fsm.If(vg.Not(out_page))(
            out_page_comp_offset(out_page_size),
            out_page_dma_offset(0),
            out_page(1)
        )
        fsm.If(out_page)(
            out_page_comp_offset(0),
            out_page_dma_offset(out_page_size),
            out_page(0)
        )

        # ReadAct and WriteOut: prev
        fsm(
            prev_row_count(row_count),
            prev_bat_count(bat_count)
        )

        # ReadAct, Comp, WriteOut: skip
        fsm.If(row_count >= self.max_row_count,
               bat_count >= self.max_bat_count)(
            skip_read_act(1)
        )

        fsm.If(row_count >= self.max_row_count,
               bat_count >= self.max_bat_count)(
            skip_comp(1)
        )

        fsm.If(skip_write_out,
               prev_row_count == 0,
               prev_bat_count == 0)(
            skip_write_out(0)
        )

        fsm.goto(state_read_act)
        fsm.If(vg.Not(skip_write_out),
               prev_row_count >= self.max_row_count,
               prev_bat_count >= self.max_bat_count).goto_next()

        # wait for last DMA write
        bt.dma_wait_write(self.maxi, fsm)


class avg_pool_serial(_pool_serial):

    def __sub_str__(self):
        sum_dtype = (' sum_dtype:%s' % self.sum_dtype.to_str()
                     if self.sum_dtype is not None else '')
        force_div = ' force_div' if self.force_div else ''

        return ''.join([_pool_serial.__sub_str__(self), sum_dtype, force_div])

    def __init__(self, value, ksize, strides, padding='SAME',
                 dtype=None, sum_dtype=None, name=None, par=1,
                 force_div=False,
                 value_ram_size=None, out_ram_size=None):

        self.sum_dtype = sum_dtype
        self.force_div = force_div
        _pool_serial.__init__(self, value, ksize, strides,
                              padding, dtype, name, par,
                              value_ram_size, out_ram_size)

    def get_pad_value(self, strm):
        return strm.Int(0)

    def get_required_substreams(self):
        ksize_col = self.ksize[-2]
        ksize_row = self.ksize[-3]
        num_vars = ksize_col * ksize_row

        act = self.args[0]
        if self.sum_dtype is not None:
            x_datawidth = self.sum_dtype.width
        else:
            x_datawidth = max(self.get_op_width(), act.get_op_width())
        x_point = act.get_op_point()
        x_signed = act.get_signed()

        substrms = [('acc_rshift_round',
                     (x_datawidth, x_point, x_signed,
                      x_datawidth, x_point, x_signed))] * self.par

        if self.force_div or num_vars & (num_vars - 1) != 0:
            y_datawidth = max(num_vars.bit_length() + 1, 2)
            y_point = 0
            y_signed = True
            substrms.extend([('div_const_frac',
                              (x_datawidth, x_point, x_signed,
                               y_datawidth, y_point, y_signed))] * self.par)

        return substrms

    def get_stream_hash(self):
        base = _pool_serial.get_stream_hash(self)
        return (base, self.sum_dtype, self.force_div)

    def pool_op(self, strm, index, var, size):
        ksize_col = self.ksize[-2]
        ksize_row = self.ksize[-3]
        num_vars = ksize_col * ksize_row

        acc = strm.substream(self.substreams[index])
        acc.to_source('x', var)
        acc.to_parameter('size', size)

        sum = acc.from_sink('sum')
        valid = acc.from_sink('valid')

        if self.force_div or num_vars & (num_vars - 1) != 0:
            rshift = 0
            acc.to_parameter('rshift', rshift)

            div = strm.substream(self.substreams[self.par + index])
            div.to_source('x', sum)
            div.to_parameter('y', num_vars)
            frac = num_vars // 2
            div.to_parameter('frac', frac)

            return div.from_sink('z'), valid

        rshift = int(math.log(num_vars, 2))
        acc.to_parameter('rshift', rshift)

        return sum, valid


class max_pool_serial(_pool_serial):

    def get_pad_value(self, strm):
        arg = self.args[0]
        datawidth = arg.get_op_width()
        value = (-1) * (1 << (datawidth - 1))
        return strm.Int(value)

    def get_required_substreams(self):
        ksize_col = self.ksize[-2]
        ksize_row = self.ksize[-3]
        num_vars = ksize_col * ksize_row

        act = self.args[0]
        width = max(self.get_op_width(), act.get_op_width())
        point = act.get_op_point()
        signed = act.get_signed()

        substrms = [('reduce_max',
                     (width, point, signed))] * self.par

        return substrms

    def pool_op(self, strm, index, var, size):
        ksize_col = self.ksize[-2]
        ksize_row = self.ksize[-3]
        num_vars = ksize_col * ksize_row

        max = strm.substream(self.substreams[index])
        max.to_source('x', var)
        max.to_parameter('size', size)

        data = max.from_sink('data')
        valid = max.from_sink('valid')

        return data, valid



from __future__ import absolute_import
from __future__ import print_function

import os
import sys

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

import nngen as ng
import veriloggen

import matrix_matmul_relu


a_shape = (15, 15)
b_shape = (15, 15)
bias_shape = None
scale_shape = None
a_dtype = ng.int32
b_dtype = ng.int32
bias_dtype = ng.int32
scale_dtype = ng.int32
c_dtype = ng.int32
rshift_mul = None
rshift_sum = None
rshift_out = None
act_func = None
par_left_col = 1
par_left_row = 1
par_out_col = 2
concur_out_col = None
stationary = 'right'
left_ram_size = None
right_ram_size = None
bias_ram_size = None
scale_ram_size = None
out_ram_size = None
axi_datawidth = 64


def test(request, silent=True):
    veriloggen.reset()

    simtype = request.config.getoption('--sim')

    rslt = matrix_matmul_relu.run(a_shape, b_shape,
                                  bias_shape, scale_shape,
                                  a_dtype, b_dtype,
                                  bias_dtype, scale_dtype,
                                  c_dtype,
                                  rshift_mul, rshift_sum, rshift_out,
                                  act_func,
                                  par_left_col, par_left_row, par_out_col,
                                  concur_out_col, stationary,
                                  left_ram_size, right_ram_size,
                                  bias_ram_size, scale_ram_size,
                                  out_ram_size,
                                  axi_datawidth, silent,
                                  filename=None, simtype=simtype,
                                  outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')

    verify_rslt = [line for line in rslt.splitlines() if line.startswith('# verify:')][0]
    assert(verify_rslt == '# verify: PASSED')


if __name__ == '__main__':
    rslt = matrix_matmul_relu.run(a_shape, b_shape,
                                  bias_shape, scale_shape,
                                  a_dtype, b_dtype,
                                  bias_dtype, scale_dtype,
                                  c_dtype,
                                  rshift_mul, rshift_sum, rshift_out,
                                  act_func,
                                  par_left_col, par_left_row, par_out_col,
                                  concur_out_col, stationary,
                                  left_ram_size, right_ram_size,
                                  bias_ram_size, scale_ram_size,
                                  out_ram_size,
                                  axi_datawidth, silent=False,
                                  filename='tmp.v',
                                  outputfile=os.path.splitext(os.path.basename(__file__))[0] + '.out')
    print(rslt)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import math
from collections import OrderedDict

import veriloggen as vg
from veriloggen.optimizer import try_optimize as optimize

import nngen.basic_types as bt
import nngen.util as util


class add(bt._ElementwiseOperator):
    """
    Returns :math:`x + y` element-wise.

    Args:
        x: A tensor.
        y: A tensor.
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Add(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class sub(bt._ElementwiseOperator):
    """
    Returns :math:`x - y` element-wise.

    Args:
        x: A tensor.
        y: A tensor.
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Sub(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class neg(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Uminus(*args)

    def __init__(self, x, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class abs(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Abs(*args)

    def __init__(self, x, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class zeros_imm(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    def op(self, strm, *args, **kwargs):
        return strm.ReinterpretCast(strm.Int(0),
                                    width=self.dtype.width,
                                    point=self.dtype.point,
                                    signed=self.dtype.signed)

    def __init__(self, shape, dtype=None, name=None, par=1):
        bt._ElementwiseOperator.__init__(self,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['shape'] = self.shape
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


def zeros_imm_like(x, dtype=None, name=None, par=1):
    shape = x.shape
    return zeros_imm(shape, dtype, name, par)


class ones_imm(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    def op(self, strm, *args, **kwargs):
        if self.dtype.point > 0:
            return strm.ReinterpretCast(strm.Int(1) << self.dtype.point,
                                        width=self.dtype.width,
                                        point=self.dtype.point,
                                        signed=self.dtype.signed)
        if self.dtype.point < 0:
            return strm.ReinterpretCast(strm.Int(1) >> self.dtype.point,
                                        width=self.dtype.width,
                                        point=self.dtype.point,
                                        signed=self.dtype.signed)

        return strm.ReinterpretCast(strm.Int(1),
                                    width=self.dtype.width,
                                    point=self.dtype.point,
                                    signed=self.dtype.signed)

    def __init__(self, shape, dtype=None, name=None, par=1):
        bt._ElementwiseOperator.__init__(self,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['shape'] = self.shape
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


def ones_imm_like(x, dtype=None, name=None, par=1):
    shape = x.shape
    return ones_imm(shape, dtype, name, par)


class full_imm(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    def __sub_str__(self):
        return ' fill_value:%d' % self.fill_value

    def get_local_control_param_values(self):
        return OrderedDict([('fill_value_cparam', self.fill_value)])

    def op(self, strm, *args, **kwargs):
        if self.dtype.point > 0:
            return strm.ReinterpretCast(self.fill_value_cparam << self.dtype.point,
                                        width=self.dtype.width,
                                        point=self.dtype.point,
                                        signed=self.dtype.signed)
        if self.dtype.point < 0:
            return strm.ReinterpretCast(self.fill_value_cparam >> self.dtype.point,
                                        width=self.dtype.width,
                                        point=self.dtype.point,
                                        signed=self.dtype.signed)

        return strm.ReinterpretCast(self.fill_value_cparam,
                                    width=self.dtype.width,
                                    point=self.dtype.point,
                                    signed=self.dtype.signed)

    def __init__(self, shape, fill_value, dtype=None, name=None, par=1):
        self.fill_value = fill_value
        bt._ElementwiseOperator.__init__(self,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['shape'] = self.shape
        kwargs['fill_value'] = self.fill_value
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


def full_imm_like(x, fill_value, dtype=None, name=None, par=1):
    shape = x.shape
    return full_imm(shape, fill_value, dtype, name, par)


class equal(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Eq(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class not_equal(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.NotEq(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class less(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.LessThan(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class less_equal(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.LessEq(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class greater(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.GreaterThan(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class greater_equal(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.GreaterEq(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class sign_binary(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Mux(args[0] > strm.Int(0), strm.Int(1), strm.Int(-1))

    def __init__(self, x, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class sign_ternary(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Mux(args[0] > strm.Int(0), strm.Int(1),
                        strm.Mux(args[0] == strm.Int(0), strm.Int(0),
                                 strm.Int(-1)))

    def __init__(self, x, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class where(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Mux(args[0], args[1], args[2])

    def __init__(self, condition, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, condition, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['condition_dtype'] = self.args[0].dtype
        kwargs['x_dtype'] = self.args[1].dtype
        kwargs['y_dtype'] = self.args[2].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class add_n(bt._ElementwiseOperator):
    """
    Adds all input tensors element-wise.

    Args:
        arg: A list of Tensor objects.
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.AddN(*args)

    def __init__(self, arg, dtype=None, name=None, par=1):
        if not isinstance(arg, (tuple, list)):
            raise TypeError('expected tuple or list')
        shape = None
        bt._ElementwiseOperator.__init__(self, *arg,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        if id(self) in memo:
            return memo[id(self)]

        import nngen.verify as verify

        name = self.__class__.__name__

        args = [arg.eval(memo, input_dict)
                for arg in self.args]

        arg_dtypes = [arg.dtype for arg in self.args]
        kwargs['arg_dtypes'] = arg_dtypes
        kwargs['dtype'] = self.dtype
        kwargs['name'] = self.name
        kwargs['par'] = self.par

        method = self.get_eval_method()
        ret = method(args, **kwargs)
        memo[id(self)] = ret

        return ret


class lshift(bt._ElementwiseOperator):
    """
    Elementwise computes the bitwise left-shift of x and y.

    Args:
        x: A tensor.
        y: A tensor.
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Sll(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class rshift(bt._ElementwiseOperator):
    """
    Elementwise computes the bitwise right-shift of x and y.

    Args:
        x: A tensor.
        y: A tensor.
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Sra(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class rshift_round(bt._ElementwiseOperator):
    """
    Elementwise computes the bitwise right-shift of x and y with rounding.

    Args:
        x: A tensor.
        y: A tensor.
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.SraRound(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class clip(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    def op(self, strm, *args, **kwargs):
        p_th, n_th = util.clip_threshold(self.dtype.width, self.dtype.signed, self.asymmetric_clip)
        p = strm.Mux(args[0] > p_th, p_th, args[0])
        n = strm.Mux(args[0] < n_th, n_th, args[0])
        return strm.Mux(args[0] >= 0, p, n)

    def __init__(self, x, asymmetric_clip=False,
                 dtype=None, name=None, par=1):

        shape = None
        bt._ElementwiseOperator.__init__(self, x,
                                         dtype=dtype, shape=shape, name=name, par=par)
        self.asymmetric_clip = asymmetric_clip

    def eval(self, memo, input_dict, **kwargs):
        kwargs['asymmetric_clip'] = self.asymmetric_clip
        kwargs['x_dtype'] = self.args[0].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class multiply(bt._ElementwiseOperator):
    """
    Returns :math:`x * y` element-wise.

    Args:
        x: A tensor.
        y: A tensor.
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    input_chainable = True
    output_chainable = True

    @staticmethod
    def op(strm, *args, **kwargs):
        return strm.Times(*args)

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class multiply_shared(multiply):
    input_chainable = True
    output_chainable = True

    def op(self, strm, *args, **kwargs):
        index = kwargs['index']
        sub = strm.substream(self.substreams[index])
        sub.to_source('x', args[0])
        sub.to_source('y', args[1])
        return sub.from_sink('z')

    def get_required_substreams(self):
        substrms = []
        substrms.extend(bt._ElementwiseOperator.get_required_substreams(self))

        x_datawidth = self.args[0].get_op_width()
        x_point = self.args[0].get_op_point()
        x_signed = self.args[0].get_signed()
        y_datawidth = self.args[1].get_op_width()
        y_point = self.args[1].get_op_point()
        y_signed = self.args[1].get_signed()
        mul_width = self.get_op_width()
        mul_point = self.get_op_point()
        mul_signed = self.get_signed()
        substrms.extend([('mul',
                          (x_datawidth, x_point, x_signed,
                           y_datawidth, y_point, y_signed,
                           mul_width, mul_point, mul_signed))] * self.par)

        return substrms


class div(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    def op(self, strm, *args, **kwargs):
        index = kwargs['index']
        sub = strm.substream(self.substreams[index])
        sub.to_source('x', args[0])
        sub.to_source('y', args[1])
        return sub.from_sink('z')

    def get_required_substreams(self):
        substrms = []
        substrms.extend(bt._ElementwiseOperator.get_required_substreams(self))

        x_datawidth = self.args[0].get_op_width()
        x_point = self.args[0].get_op_point()
        x_signed = self.args[0].get_signed()
        y_datawidth = self.args[1].get_op_width()
        y_point = self.args[1].get_op_point()
        y_signed = self.args[1].get_signed()
        div_width = self.get_op_width()
        div_point = self.get_op_point()
        div_signed = self.get_signed()

        substrms.extend([('div',
                          (x_datawidth, x_point, x_signed,
                           y_datawidth, y_point, y_signed,
                           div_width, div_point, div_signed))] * self.par)

        return substrms

    def __init__(self, x, y, dtype=None, name=None, par=1):
        shape = None
        bt._ElementwiseOperator.__init__(self, x, y,
                                         dtype=dtype, shape=shape, name=name, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class multiply_add_rshift_clip(bt._ElementwiseOperator):
    input_chainable = True
    output_chainable = True

    def op(self, strm, *args, **kwargs):
        x_datawidth = self.args[0].get_op_width()
        x_point = self.args[0].get_op_point()
        x_signed = self.args[0].get_signed()
        y_datawidth = self.args[1].get_op_width()
        y_point = self.args[1].get_op_point()
        y_signed = self.args[1].get_signed()
        z_datawidth = self.args[2].get_op_width()
        z_point = self.args[2].get_op_point()
        z_signed = self.args[2].get_signed()

        madd = strm.Madd(args[0], args[1], args[2])

        if self.sum_dtype is not None:
            madd.width = self.sum_dtype.width
            madd.signed = self.sum_dtype.signed
        else:
            madd.width = max(x_datawidth + y_datawidth, z_datawidth)
            madd.signed = self.dtype.signed

        if self.sum_dtype is not None and madd.point != self.sum_dtype.point:
            madd = strm.Cast(madd, point=self.sum_dtype.point)

        sra = strm.Sra(madd, args[3])

        p_th, n_th = util.clip_threshold(self.dtype.width, self.dtype.signed, self.asymmetric_clip)
        p = strm.Mux(sra > p_th, p_th, sra)
        n = strm.Mux(sra < n_th, n_th, sra)

        return strm.Mux(sra >= 0, p, n)

    def __init__(self, x, y, z, shamt, asymmetric_clip=False,
                 dtype=None, sum_dtype=None, name=None, par=1):

        shape = None
        bt._ElementwiseOperator.__init__(self, x, y, z, shamt,
                                         dtype=dtype, shape=shape, name=name, par=par)
        self.asymmetric_clip = asymmetric_clip
        self.sum_dtype = sum_dtype

    def eval(self, memo, input_dict, **kwargs):
        kwargs['asymmetric_clip'] = self.asymmetric_clip
        kwargs['x_dtype'] = self.args[0].dtype
        kwargs['y_dtype'] = self.args[1].dtype
        kwargs['z_dtype'] = self.args[2].dtype
        kwargs['shamt_dtype'] = self.args[3].dtype
        return bt._ElementwiseOperator.eval(self, memo, input_dict, **kwargs)


class _reduce_op(bt._ReductionOperator):

    input_chainable = True
    output_chainable = False

    def __init__(self, input_tensor,
                 axis=None, keep_dims=False, dtype=None, name=None, par=1):

        rank = bt.get_rank(input_tensor.shape)
        axis = util.to_axis(axis, rank)
        shape = util.to_reduce_shape(input_tensor.shape, axis, keep_dims)

        bt._ReductionOperator.__init__(self, input_tensor,
                                       dtype=dtype, shape=shape, name=name,
                                       axis=axis, keep_dims=keep_dims, par=par)

    def eval(self, memo, input_dict, **kwargs):
        kwargs['input_tensor_dtype'] = self.args[0].dtype
        return bt._ReductionOperator.eval(self, memo, input_dict, **kwargs)


class reduce_sum(_reduce_op):
    """
    Computes the sum of elements across dimensions of a tensor.

    Args:
        input_tensor: A tensor.
        axis: The dimensions to reduce (optional).
        keep_dims: If true, retains reduced dimensions with length 1 (optional).
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    @staticmethod
    def reduce_op(strm, value, size, par_offset, **kwargs):
        data, valid = strm.ReduceAddValid(value, size, **kwargs)
        return (data,), valid

    @staticmethod
    def carry_op(strm, *args, **kwargs):
        return strm.Add(*args)


class reduce_max(_reduce_op):
    """
    Returns the maximum value of elements across dimensions of a tensor.

    Args:
        input_tensor: A tensor.
        axis: The dimensions to reduce (optional).
        keep_dims: If true, retains reduced dimensions with length 1 (optional).
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    def reduce_op(self, strm, value, size, par_offset, **kwargs):
        kwargs['initval'] = self.carry_default_values[0]
        data, valid = strm.ReduceMaxValid(value, size, **kwargs)
        return (data,), valid

    @staticmethod
    def carry_op(strm, *args, **kwargs):
        return strm.Max(*args)

    def __init__(self, input_tensor,
                 dtype=None, shape=None, name=None,
                 axis=None, keep_dims=False, par=1):

        _reduce_op.__init__(self, input_tensor,
                            axis=axis, keep_dims=keep_dims, dtype=dtype, name=name, par=par)

        if self.get_signed():
            point = self.get_op_point()
            if point > 0:
                default_value = -1 * ((2 ** (self.dtype.width - 1)) >> point)
            else:
                default_value = -1 * ((2 ** (self.dtype.width - 1)) << -point)
        else:
            default_value = 0

        self.carry_default_values = [default_value]


class reduce_min(_reduce_op):
    """
    Returns the minimum value of elements across dimensions of a tensor.

    Args:
        input_tensor: A tensor.
        axis: The dimensions to reduce (optional).
        keep_dims: If true, retains reduced dimensions with length 1 (optional).
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    def reduce_op(self, strm, value, size, par_offset, **kwargs):
        kwargs['initval'] = self.carry_default_values[0]
        data, valid = strm.ReduceMinValid(value, size, **kwargs)
        return (data,), valid

    @staticmethod
    def carry_op(strm, *args, **kwargs):
        return strm.Min(*args)

    def __init__(self, input_tensor,
                 dtype=None, shape=None, name=None,
                 axis=None, keep_dims=False, par=1):

        _reduce_op.__init__(self, input_tensor,
                            axis=axis, keep_dims=keep_dims, dtype=dtype, name=name, par=par)

        if self.get_signed():
            point = self.get_op_point()
            if point > 0:
                default_value = (2 ** (self.dtype.width - 1) - 1) >> point
            else:
                default_value = (2 ** (self.dtype.width - 1) - 1) << -point
        else:
            if point > 0:
                default_value = (2 ** self.dtype.width - 1) >> point
            else:
                default_value = (2 ** self.dtype.width - 1) << -point

        self.carry_default_values = (default_value,)


class _arg_op(_reduce_op):

    carry_default_values = (0, 0)

    def __init__(self, input_tensor,
                 axis=None, keep_dims=False, dtype=None, name=None, par=1):

        if isinstance(axis, (tuple, list)) and len(axis) > 1:
            raise ValueError('size of axis must be 1.')

        # Because get_stream_reduce_output() cannot calculate
        # the flatten index value for a unflatten value,
        # input value must be flatten when axis is None.
        if axis is None and len(input_tensor.shape) > 1:
            input_tensor = reshape(input_tensor, [-1])

        _reduce_op.__init__(self, input_tensor,
                            axis=axis, keep_dims=keep_dims, dtype=dtype, par=par)


class argmax(_arg_op):
    """
    Returns the index of the maximum value of elements across dimensions of a tensor.

    Args:
        input_tensor: A tensor.
        axis: The dimensions to reduce (optional).
        keep_dims: If true, retains reduced dimensions with length 1 (optional).
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    def reduce_op(self, strm, value, size, par_offset, **kwargs):
        kwargs['initval'] = self.carry_default_values[1]
        index, data, valid = strm.ReduceArgMaxValid(value, size, **kwargs)
        shamt = math.ceil(math.log2(self.par))
        index = index << shamt
        index.latency = 0
        index = index + par_offset
        index.latency = 0
        return (index, data), valid

    def gather_op(self, strm, indexes, values, **kwargs):
        ret_index = indexes[0]
        ret_value = values[0]
        for index, value in zip(indexes[1:], values[1:]):
            ret_index = strm.Mux(ret_value < value, index, ret_index)
            ret_index.latency = 0
            ret_value = strm.Mux(ret_value < value, value, ret_value)
            ret_value.latency = 0
        return ret_index, ret_value

    def __init__(self, input_tensor,
                 axis=None, keep_dims=False, dtype=None, name=None, par=1):

        _arg_op.__init__(self, input_tensor,
                         axis=axis, keep_dims=keep_dims, dtype=dtype, name=name, par=par)

        if self.get_signed():
            point = self.get_op_point()
            if point > 0:
                default_value = -1 * ((2 ** (self.dtype.width - 1)) >> point)
            else:
                default_value = -1 * ((2 ** (self.dtype.width - 1)) << -point)
        else:
            default_value = 0

        self.carry_default_values = (0, default_value)


class argmin(_arg_op):
    """
    Returns the index of the minimum value of elements across dimensions of a tensor.

    Args:
        input_tensor: A tensor.
        axis: The dimensions to reduce (optional).
        keep_dims: If true, retains reduced dimensions with length 1 (optional).
        dtype: Output data type (optional).
        name: A name for the operation (optional).
        par: The number of parallel operations (optional).
    """

    def reduce_op(self, strm, value, size, par_offset, **kwargs):
        kwargs['initval'] = self.carry_default_values[1]
        index, data, valid = strm.ReduceArgMinValid(value, size, **kwargs)
        shamt = math.ceil(math.log2(self.par))
        index = index << shamt
        index.latency = 0
        index = index + par_offset
        index.latency = 0
        return (index, data), valid

    def gather_op(self, strm, indexes, values, **kwargs):
        ret_index = indexes[0]
        ret_value = values[0]
        for index, value in zip(indexes[1:], values[1:]):
            ret_index = strm.Mux(ret_value < value, ret_index, index)
            ret_index.latency = 0
            ret_value = strm.Mux(ret_value < value, ret_value, value)
            ret_value.latency = 0
        return ret_index, ret_value

    def __init__(self, input_tensor,
                 axis=None, keep_dims=False, dtype=None, name=None, par=1):

        _arg_op.__init__(self, input_tensor,
                         axis=axis, keep_dims=keep_dims, dtype=dtype, name=name, par=par)

        if self.get_signed():
            point = self.get_op_point()
            if point > 0:
                default_value = (2 ** (self.dtype.width - 1) - 1) >> point
            else:
                default_value = (2 ** (self.dtype.width - 1) - 1) << -point
        else:
            if point > 0:
                default_value = (2 ** self.dtype.width - 1) >> point
            else:
                default_value = (2 ** self.dtype.width - 1) << -point

        self.carry_default_values = (0, default_value)


class _reshape(bt._Reshape):
    pass


class _lazy_reshape(bt._LazyReshape):
    pass


def reshape(tensor, shape, dtype=None, name=None):
    """
    Reshapes a tensor.

    Given tensor, this operation returns a tensor \
    that has the same values as tensor with shape shape.

    Args:
        tensor: A tensor.
        shape: A tensor. Defines the shape of the output tensor.
        dtype: Output data type (optional).
        name: A name for the operation (optional).
    """
    return _lazy_reshape(tensor, shape, dtype, name)


def cast(x, dtype, name=None):
    if x.dtype.wordsize == dtype.wordsize:
        return bt._View(x, x.shape, dtype, name)

    return _reshape(x, x.shape, dtype, name)


def expand_dims(input, axis, name=None):
    """
    Inserts a dimension of 1 into a tensor's shape.

    Given a tensor input, this operation inserts a dimension \
    of 1 at the dimension index axis of input's shape. \
    The dimension index axis starts at zero; \
    if you specify a negative number for axis it is counted backward from the end.

    Args:
        input: A Tensor.
    """
    shape = input.shape
    rank = bt.get_rank(shape)
    axis = util.to_axis(axis, rank)[0]

    new_shape = []
    new_shape.extend(shape[:axis])
    new_shape.append(1)
    new_shape.extend(shape[axis:])

    if axis == rank:
        return _lazy_reshape(input, new_shape)

    return bt._View(input, shape=new_shape, name=name)


class transpose(bt._Operator):
    """
    Transposes a. Permutes the dimensions according to perm.

    Args:
        a: A Tensor.
        perm: A permutation of the dimensions of a.
        dtype: Output data type (optional).
        name: A name for the operation (optional).
    """

    input_chainable = False
    output_chainable = False
    thread_cachable = False

    def __sub_str__(self):
        perm = ' perm:%s' % str(tuple(self.transpose_perm))
        onnx_perm = (' onnx_perm:%s' % str(tuple(self.transpose_onnx_perm))
                     if self.transpose_onnx_perm is not None else '')
        return '%s%s' % (perm, onnx_perm)

    def __init__(self, a, perm=None, dtype=None, name=None):

        if bt.get_rank(a.shape) == 1:
            a_shape = tuple([1, a.shape[0]])
        else:
            a_shape = a.shape

        if perm is None:
            perm = tuple(reversed(range(bt.get_rank(a_shape))))

        self.transpose_perm = perm
        self.transpose_onnx_perm = None

        shape = []
        for p in perm:
            shape.append(a_shape[p])
        shape = tuple(shape)

        bt._Operator.__init__(self, a,
                              dtype=dtype, shape=shape, name=name)
        self.implicit = False

    def attribute(self):
        pass

    def get_required_rams(self):
        # burst read, scatter write
        min_size = self.args[0].shape[-1]
        input_width = self.args[0].get_ram_width()
        output_width = self.get_ram_width()

        inputs = [(input_width, min_size)]
        outputs = [(output_width, min_size)]
        temps = []
        return inputs, outputs, temps

    def get_stream_func(self):
        return None

    def get_control_param_values(self):
        return OrderedDict()

    def control_sequence(self, fsm):
        arg = self.args[0]
        ram = self.input_rams[0]

        shape = self.get_aligned_shape()
        arg_shape = arg.get_aligned_shape()

        # burst read, scatter write
        write_order = list(reversed([self.transpose_perm.index(i)
                                     for i in range(len(shape))]))
        write_pattern = bt.shape_to_pattern(shape, write_order)

        read_offset = self.m.TmpReg(self.maxi.addrwidth, initval=0)
        write_offsets = [self.m.TmpReg(self.maxi.addrwidth, initval=0)
                         for _ in write_pattern]
        write_all_offset = self.objaddr
        for write_offset in write_offsets:
            write_all_offset += write_offset

        read_counts = [self.m.TmpReg(self.maxi.addrwidth, initval=0)
                       for _ in write_pattern]

        # initialize
        fsm(
            read_offset(0),
            [write_offset(0) for write_offset in write_offsets],
            [read_count(0) for read_count in read_counts]
        )
        fsm.goto_next()

        # DMA read
        read_state = fsm.current

        laddr = 0
        gaddr = self.arg_objaddrs[0] + read_offset
        read_size = arg_shape[-1]

        bt.bus_lock(self.maxi, fsm)
        bt.dma_read(self.maxi, fsm, ram, laddr, gaddr, read_size)
        bt.bus_unlock(self.maxi, fsm)

        # read-modify-write
        modify_state = fsm.current

        laddr = read_counts[0]
        gaddr = write_all_offset

        bt.read_modify_write(self.m, fsm, self.maxi,
                             ram, self.output_rams[0],
                             laddr, gaddr)

        prev_done = 1
        for (read_count, maxval,
             write_offset,
             (out_size, out_stride)) in zip(read_counts, reversed(arg_shape),
                                            write_offsets,
                                            write_pattern):
            fsm.If(prev_done)(
                read_count.inc(),
                write_offset.add(
                    optimize(bt.to_byte(out_stride * self.get_ram_width())))
            )
            fsm.If(prev_done, read_count == maxval - 1)(
                read_count(0),
                write_offset(0)
            )
            prev_done = vg.Ands(prev_done, (read_count == maxval - 1))

        fsm.If(laddr == read_size - 1)(
            read_offset.add(
                optimize(bt.to_byte(read_size * arg.get_ram_width())))
        )
        fsm.If(laddr < read_size - 1).goto(modify_state)
        fsm.If(laddr == read_size - 1).goto(read_state)
        fsm.If(prev_done).goto_next()

    def get_layout(self):
        if self.layout is not None:
            return self.layout

        orig_layout = self.args[0].get_layout()
        if orig_layout is None:
            return None

        new_layout = [l for l in orig_layout]
        for i, p in enumerate(self.transpose_perm):
            new_layout[i] = orig_layout[p]
        return ''.join(new_layout)

    def get_onnx_layout(self):
        if self.onnx_layout is not None:
            return self.onnx_layout

        orig_onnx_layout = self.args[0].get_onnx_layout()
        if orig_onnx_layout is None:
            return None

        new_onnx_layout = [l for l in orig_onnx_layout]
        for i, p in enumerate(self.transpose_onnx_perm):
            new_onnx_layout[i] = orig_onnx_layout[p]
        return ''.join(new_onnx_layout)

    def eval(self, memo, input_dict, **kwargs):
        if id(self) in memo:
            return memo[id(self)]

        import nngen.verify as verify

        name = self.__class__.__name__

        args = [arg.eval(memo, input_dict)
                for arg in self.args]

        kwargs['perm'] = self.transpose_perm
        kwargs['dtype'] = self.dtype
        kwargs['name'] = self.name

        method = self.get_eval_method()
        ret = method(*args, **kwargs)
        memo[id(self)] = ret

        return ret



from __future__ import absolute_import
from __future__ import print_function

from . import conv2d


class ternary_weight_conv2d(conv2d):

    def get_required_substreams(self):
        if self.mul_dtype is not None:
            mul_width = self.mul_dtype.width
        else:
            mul_width = self.get_op_width()

        if self.mul_dtype is not None:
            mul_point = self.mul_dtype.point
        else:
            mul_point = self.get_op_point()

        if self.mul_dtype is not None:
            mul_signed = self.mul_dtype.signed
        else:
            mul_signed = self.get_signed()

        if self.sum_dtype is not None:
            sum_width = self.sum_dtype.width
        else:
            sum_width = self.get_op_width()

        if self.sum_dtype is not None:
            sum_point = self.sum_dtype.point
        else:
            sum_point = self.get_op_point()

        if self.sum_dtype is not None:
            sum_signed = self.sum_dtype.signed
        else:
            sum_signed = self.get_signed()

        filter_num_col = self.filter_shape[-2]
        filter_num_row = self.filter_shape[-3]
        num_macs = filter_num_col * filter_num_row

        x_datawidth = self.args[0].get_op_width()
        x_point = self.args[0].get_op_point()
        x_signed = self.args[0].get_signed()
        y_datawidth = self.args[1].get_op_width()
        y_point = self.args[1].get_op_point()
        y_signed = self.args[1].get_signed()

        if y_datawidth != 2:
            raise ValueError('ternary weight must be 2-bit.')

        if y_point != 0:
            raise ValueError('ternary weight must be int, not fixed.')

        if not y_signed:
            raise ValueError('ternary weight must be signed.')

        args = (x_datawidth, x_point, x_signed,
                y_datawidth, y_point, y_signed,
                mul_width, mul_point, mul_signed)

        mulname = 'updown_mask_rshift'
        substrms = [(mulname, args)] * (num_macs *
                                        self.par_ich * self.par_och)

        substrms.extend([('add_tree',
                          (sum_width, sum_point, sum_signed,
                           num_macs * self.par_ich))] * self.par_och)
        substrms.extend([('acc_rshift_round_frac',
                          (sum_width, sum_point, sum_signed,
                           sum_width, sum_point, sum_signed))] * self.par_och)
        substrms.extend([('mul_rshift',
                          (sum_width, sum_point, sum_signed,
                           sum_width, sum_point, sum_signed))] * self.par_och)

        return substrms



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np

import nngen.operator as operator
import nngen.dtype_list as dtype_list

from . import util
from . import conv
from . import gemm


def BatchNormalization(visitor, node, act_func=None):

    srcs = []
    for src in node.input[1:]:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    # expand the expression
    a_value = srcs[0].value
    b_value = srcs[1].value
    mean_value = srcs[2].value
    variance_value = srcs[3].value
    epsilon = node.attribute[0].f
    div = np.sqrt(variance_value + epsilon)

    scale_value = a_value / div
    bias_value = b_value - a_value * mean_value / div

    # input
    node_name = util.get_name(node)

    src_name = node.input[0]
    src_node = util.search_node_from_model(visitor.model, src_name)

    if (not visitor.disable_fusion and
            src_node.op_type == 'Conv' and len(visitor.consumers[src_name]) == 1):

        src_op = conv.Conv(visitor, src_node,
                           batchnorm_scale=scale_value,
                           batchnorm_bias=bias_value,
                           act_func=act_func)
        visitor.operators[node_name] = src_op
        return src_op

    if (not visitor.disable_fusion and
            src_node.op_type == 'Gemm' and len(visitor.consumers[src_name]) == 1):

        src_op = gemm.Gemm(visitor, src_node,
                           batchnorm_scale=scale_value,
                           batchnorm_bias=bias_value,
                           act_func=act_func)
        visitor.operators[node_name] = src_op
        return src_op

    if src_node.op_type == 'Unsqueeze' and len(visitor.consumers[src_name]) == 1:

        src_src_name = src_node.input[0]
        src_src_node = util.search_node_from_model(visitor.model, src_src_name)

        if (not visitor.disable_fusion and
                src_src_node.op_type == 'Gemm' and
                len(visitor.consumers[src_src_name]) == 1):

            src_src_op = gemm.Gemm(visitor, src_src_node,
                                   batchnorm_scale=scale_value,
                                   batchnorm_bias=bias_value,
                                   act_func=act_func)
            visitor.operators[node_name] = src_src_op
            return src_src_op

    input = visitor.visit(node.input[0])
    scale = srcs[0]
    bias = srcs[1]

    scale.value = scale_value
    bias.value = bias_value

    scale.dtype = visitor.default_scale_dtype
    bias.dtype = visitor.default_bias_dtype

    shamt_shape = (1,)
    shamt_dtype = dtype_list.dtype_int(scale.dtype.width, signed=True)
    shamt = operator.full_imm(shamt_shape, 0, shamt_dtype)

    out = operator.normalize(input, scale, bias, shamt)

    if act_func is not None:
        out = act_func(out)

    return out



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np

import nngen.basic_types as bt
import nngen.operator as operator

from . import util


def Cast(visitor, node):

    srcs = []

    for src in node.input:
        src_obj = visitor.visit(src)
        srcs.append(src_obj)

    input = srcs[0]

    if not isinstance(input, bt._Numeric):
        return input

    name = util.get_name(node)

    if name in visitor.value_dtypes:
        dtype = visitor.value_dtypes[name]
    else:
        dtype = visitor.default_operator_dtype

    c = operator.cast(input, dtype)
    c.layout = input.get_layout()
    c.onnx_layout = input.get_onnx_layout()

    return c



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np


def pad(value, padding,
        dtype=None, name=None, par=1,
        value_ram_size=None, out_ram_size=None,
        value_dtype=None):

    ksize_row = 1
    ksize_col = 1
    stride_row = 1
    stride_col = 1

    if isinstance(padding, str) and padding == 'SAME':
        pad_col, pad_col_left, pad_col_right = util.pad_size_split(
            value.shape[2], ksize_col, stride_col)
        pad_row, pad_row_top, pad_row_bottom = util.pad_size_split(
            value.shape[1], ksize_row, stride_row)

    elif isinstance(padding, str) and padding == 'VALID':
        pad_col, pad_col_left, pad_col_right = 0, 0, 0
        pad_row, pad_row_top, pad_row_bottom = 0, 0, 0

    elif isinstance(padding, int):
        pad_col, pad_col_left, pad_col_right = padding * 2, padding, padding
        pad_row, pad_row_top, pad_row_bottom = padding * 2, padding, padding

    elif isinstance(padding, (tuple, list)):
        pad_col, pad_col_left, pad_col_right = padding[2] + padding[3], padding[2], padding[3]
        pad_row, pad_row_top, pad_row_bottom = padding[0] + padding[1], padding[0], padding[1]

    else:
        raise ValueError("padding options must be 'SAME', 'VALID', int, tuple, or list.")

    out = np.pad(value, [(0, 0),
                         (pad_row_top, pad_row_bottom),
                         (pad_col_left, pad_col_right),
                         (0, 0)], 'constant')

    return out



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import math

base_order = ['int', 'fixed']


class dtype_info(object):

    def __repr__(self):
        return '<dtype %s%s%d%s>' % ('' if self.signed else 'u',
                                     self.base, self.width,
                                     '' if self.base == 'int' else '_%d' % self.point)

    def to_str(self):
        return '%s%s%d%s' % ('' if self.signed else 'u',
                             self.base, self.width,
                             '' if self.base == 'int' else '_%d' % self.point)

    def __hash__(self):
        return hash((self.base, self.width, self.point, self.signed))

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        return ((self.base, self.width, self.point, self.signed) ==
                (other.base, other.width, other.point, other.signed))

    def __init__(self, base, width, point=0, signed=True):
        if base != 'fixed' and point != 0:
            raise ValueError("point must be 0 for dtype '%s'" % base)

        self.base = base
        self.width = width
        self.point = point
        self.signed = signed

    @property
    def wordsize(self):
        return int(math.ceil(self.width / 8))


def dtype_int(width=None, signed=True):
    base = 'int'
    if width is None:
        width = 32
    point = 0
    return dtype_info(base, width, point, signed)


def dtype_fixed(width=None, point=None, signed=True):
    base = 'fixed'
    if width is None:
        width = 32
    if point is None:
        point = 0
    return dtype_info(base, width, point, signed)


def get_max_dtype(*args):
    dtype = None
    for arg in args:
        if arg.dtype is None:
            continue

        if dtype is None:
            dtype = arg.dtype
            continue

        base = base_order[max(base_order.index(arg.dtype.base),
                              base_order.index(dtype.base))]
        width = max(arg.dtype.width, dtype.width)
        point = max(arg.dtype.point, dtype.point)
        signed = arg.dtype.signed or dtype.signed
        dtype = dtype_info(base, width, point, signed)

    return dtype


int64 = dtype_int(64, True)
int32 = dtype_int(32, True)
int16 = dtype_int(16, True)
int8 = dtype_int(8, True)
int4 = dtype_int(4, True)
int2 = dtype_int(2, True)
uint64 = dtype_int(64, False)
uint32 = dtype_int(32, False)
uint16 = dtype_int(16, False)
uint8 = dtype_int(8, False)
uint4 = dtype_int(4, False)
uint2 = dtype_int(2, False)
uint1 = dtype_int(1, False)

fixed32_16 = dtype_fixed(32, 16, True)
fixed16_8 = dtype_fixed(16, 8, True)
fixed8_4 = dtype_fixed(8, 4, True)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

from .basic import *
from .exp import exp
from .relu import relu, relu6
from .leaky_relu import leaky_relu, get_leaky_relu_op, leaky_relu_base
from .sigmoid import sigmoid
from .matmul import matmul
from .conv2d import conv2d
from .log_weight_conv2d import log_weight_conv2d
from .binary_weight_conv2d import binary_weight_conv2d
from .ternary_weight_conv2d import ternary_weight_conv2d
from .pool import avg_pool, max_pool
from .pool_serial import avg_pool_serial, max_pool_serial
from .extern import extern
from .concat import concat
from .slice_ import slice_
from .upsampling2d import upsampling2d
from .pad import pad
from .normalize import *
